#
# -*- coding: utf-8 -*-
#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2003-2006  Donald N. Allingham
# Copyright (C) 2021-2022  Bora Atıcı
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Written by Bora Atıcı <boratici.acc@gmail.com>, 2021
#
# --------------------------------------------------------------------------------
#
# Turkish relationship and kinship names
#
# --------------------------------------------------------------------------------
# kan bağı(blood ties)
# - haminne(parent's grandmother)
# - dede(grandfather), nine(grandmother) -> babaanne(father's mother),
#                                        -> anneanne(mother's mother)
# - baba/ata(father), anne/ana(mother)
# - kardeş(sibling), erkek kardeş(brother), kız kardeş(sister)
#                    abi(big brother)     , abla(big sister)
# - torun(grandchild)
# - -> amca(father's brother/uncle), dayı(mother's brother/uncle)
#   -> hala(father's sister/aunt)  , teyze(mother's sister/aunt)
# - kuzen/böle(only uncle and aunt's children/cousin)
# - yeğen(only sibling's children/nephew-niece)
#
# kayın(in-law)
# - kaynata/kayınbaba(father-in-law), kaynana/kayınvalide(mother-in-law)
# - -> kayınbirader/kayın(only spouse's brother/brother-in-law),
#   -> -> görümce(only husband's sister/sister-in-law)
#      -> baldız(only wife's sister/sister-in-law)
# - damat
# - gelin
# - enişte
# - yenge
# - bacanak
# - görümce
# - elti
# - babalık
# - analık
# - oğulluk
# - dünür
# - ebeveyn
# and much more. It will be added with future updates.
# --------------------------------------------------------------------------------
"""
Turkish-specific classes for relationships.
"""
# --------------------------------------------------------------------------------
#
# Gramps modules
#
# --------------------------------------------------------------------------------
from operator import le
from gramps.gen.lib import Person
import gramps.gen.relationship

# --------------------------------------------------------------------------------
#
# Shared constants
#
# --------------------------------------------------------------------------------
_level_name = [
    "",
    "",
    "ikinci derece",
    "üçüncü derece",
    "dördüncü derece",
    "beşinci derece",
    "altıncı derece",
    "yedinci derece",
    "sekizinci derece",
    "dokuzuncu derece",
    "onuncu derece",
    "on birinci",
    "on ikinci",
    "on üçüncü",
    "on dördüncü",
    "on beşinci",
    "on altıncı",
    "on yedinci",
    "on sekizinci",
    "on dokuzuncu",
    "yirminci",
    "yirmi birinci",
    "yirmi ikinci",
    "yirmi üçüncü",
    "yirmi dördüncü",
    "yirmi beşinci",
    "yirmi altıncı",
    "yirmi yedinci",
    "yirmi sekizinci",
    "yirmi dokuzuncu",
    "otuzuncu",
    "otuz birinci",
    "otuz ikinci",
    "otuz üçüncü",
    "otuz dördüncü",
    "otuz beşinci",
    "otuz altıncı",
    "otuz yedinci",
    "otuz sekizinci",
    "otuz dokuzuncu",
    "kırkıncı",
    "kırk birinci",
]

_level_name_male = _level_name

_level_name_female = _level_name

# --------------------------------------------------------------------------------
#
# Relationship levels
#
# --------------------------------------------------------------------------------
_parent_level = ["", "", "ebeveynlerinin", "büyük ebeveynlerinin"]

_son_level = [
    "",
    "oğlu",
    "torunu",
    "torununun oğlu",
    "torununun torunu",
    "torununun torununun oğlu",
    "torununun torununun torunu",
]

_daughter_level = [
    "",
    "kızı",
    "torunu",
    "torununun kızı",
    "torununun torunu",
    "torununun torununun kızı",
    "torununun torununun torunu",
]

_child_level = [
    "",
    "çocuğu",
    "torunu",
    "torununun çocuğu",
    "torununun torunu",
    "torununun torununun çocuğu",
    "torununun torununun torunu",
]

_father_level = [
    "",
    "babası",
    "dedesi",
    "büyük dedesi",
    "dedesinin dedesi",
    "dedesinin dedesinin babası",
    "dedesinin dedesinin dedesi",
]

_mother_level = [
    "",
    "annesi",
    "ninesi",
    "haminnesi",
    "büyük haminnesi",
    "ninesinin ninesinin annesi",
    "ninesinin ninesinin ninesi",
]

_uncle_level = [
    "",
    "erkek kardeş",
    "amca",
    "büyük amca",
]

_aunt_level = [
    "",
    "kız kardeş",
    "hala",
    "büyük hala",
]

# --------------------------------------------------------------------------------
#
# Relationship Classes
#
# --------------------------------------------------------------------------------


class RelationshipCalculator(gramps.gen.relationship.RelationshipCalculator):
    """
    RelationshipCalculator Class
    """

    def __init__(self):
        gramps.gen.relationship.RelationshipCalculator.__init__(self)

    def get_sword_distaff(self, level, reltocommon, gender=Person.UNKNOWN):
        """
        Generate relationships 'by male line' or 'by female line',
        specific for Turkish
        """
        if level <= 1:
            return ""

        by_line = ""
        for g in reltocommon:
            if by_line and by_line != g:
                by_line = ""
                break
            by_line = g

        if by_line == self.REL_FATHER and gender == Person.MALE:
            # By male line
            return "(erkek hattından) "
        elif by_line == self.REL_MOTHER and gender == Person.FEMALE:
            # By male line
            return "(kadın hattından) "
        elif reltocommon[0] == self.REL_FATHER:
            # From father's side
            return "(baba tarafından) "
        elif reltocommon[0] == self.REL_MOTHER:
            # From mother's side
            return "(anne tarafından) "
        else:
            return ""

    def get_son(self, level, inlaw=""):
        """
        Provides information how much male descendant is related to this person
        """
        # Define if the person is natural relative or in law
        if inlaw == "":
            t_inlaw = ""
        else:
            t_inlaw = "kayın "

        if level >= 0 and level < len(_son_level):
            return t_inlaw + _son_level[level]
        # else:
        # return t_inlaw + "%d" % (level + 1)

    def get_daughter(self, level, inlaw=""):
        """
        Provides information how much female descendant is related to this person
        """
        # Define if the person is natural relative or in law
        if inlaw == "":
            t_inlaw = ""
        else:
            t_inlaw = "kayın "

        if level >= 0 and level < len(_daughter_level):
            return t_inlaw + _daughter_level[level]
        # else:
        # return t_inlaw + "%d" % (level + 1)

    def get_child_unknown(self, level, inlaw=""):
        """
        Provides information how much descendant of unknown gender is related to this person
        """
        # Define if the person is natural relative or in law
        if inlaw == "":
            t_inlaw = ""
        else:
            t_inlaw = "kayın "

        if level >= 0 and level < len(_child_level):
            return t_inlaw + _child_level[level]
        # else:
        # return t_inlaw + "%d" % (level + 1)

    def get_father(self, level, reltocommon, inlaw=""):
        """
        Provides information how much male ancestor (for example father)
        is related to the given person
        """
        if inlaw == "":
            t_inlaw = ""
        else:
            t_inlaw = "kayın "

        if level >= 0 and level < len(_father_level):
            # If you know exact name of relationship - use it
            if level == 1:
                # Father
                return t_inlaw + _father_level[level]
            else:
                # Grandfather, Greatgrandfather, Greatgreatgrandfather
                return (
                    t_inlaw
                    + self.get_sword_distaff(level, reltocommon, Person.MALE)
                    + _father_level[level]
                )
        else:
            # For deep generations
            return (
                t_inlaw
                + "%s dedesi" % _level_name[level]
                + self.get_sword_distaff(level, reltocommon, Person.MALE)
            )

    def get_mother(self, level, reltocommon, inlaw=""):
        """
        Provides information how much female ancestor (for example mother)
        is related to the given person
        """
        if inlaw == "":
            t_inlaw = ""
        else:
            t_inlaw = "kayın "

        if level >= 0 and level < len(_mother_level):
            # If you know exact name of relationship - use it
            if level == 1:
                # Mother
                return t_inlaw + _mother_level[level]

            else:
                # Grandmother, Greatgrandmother, Greatgreatgrandmother
                return (
                    t_inlaw
                    + self.get_sword_distaff(level, reltocommon, Person.FEMALE)
                    + _mother_level[level]
                )
        else:
            # For deep generations
            #
            return (
                t_inlaw
                + "%s ninesi" % _level_name[level]
                + self.get_sword_distaff(level, reltocommon, Person.FEMALE)
            )

    def get_parent_unknown(self, level, reltocommon, inlaw=""):
        """
        Provides information how much an ancestor of unknown gender
        is related to the given person (unknown sex)
        """
        if inlaw == "":
            t_inlaw = ""
        else:
            t_inlaw = "kayın "

        if level == 1:
            return t_inlaw + "babası/annesi"
        else:
            return (
                t_inlaw
                + "%s dedesi/ninesi" % _level_name[level]
                + self.get_sword_distaff(level, reltocommon)
            )

    def get_uncle(self, level, reltocommon, inlaw=""):
        """
        Return uncle generation name
        """
        if reltocommon[0] == self.REL_FATHER:
            # from Father
            if level < len(_uncle_level):
                if level == 2:
                    return "amcası"
                elif level == 3:
                    return "(baba tarafından) büyük amcası"
                elif level == 4:
                    return "(baba tarafından) büyük büyük amcası"
            else:
                return "(baba tarafından) " + (_level_name[level - 1]) + " amcası"
        elif reltocommon[0] == self.REL_MOTHER:
            # from Mother
            if level < len(_uncle_level):
                if level == 2:
                    return "dayısı"
                elif level == 3:
                    return "(anne tarafından) büyük dayısı"
                elif level == 4:
                    return "(anne tarafından) büyük büyük dayısı"
            else:
                return "(anne tarafından) " + (_level_name[level - 1]) + " dayısı"
        else:
            if level == 2:
                return "amcası/dayısı"
            else:
                return (_level_name[level - 1]) + " amcası/dayısı"

    def get_aunt(self, level, reltocommon, inlaw=""):
        """
        Return aunt generation name
        """
        if reltocommon[0] == self.REL_FATHER:
            # from Father
            if level < len(_aunt_level):
                if level == 2:
                    return "halası"
                elif level == 3:
                    return "(baba tarafından) büyük halası"
                elif level == 4:
                    return "(baba tarafından) büyük büyük halası"
            else:
                return "(baba tarafından) %s halası" % _level_name[level - 1]
        elif reltocommon[0] == self.REL_MOTHER:
            # from Mother
            if level < len(_aunt_level):
                if level == 2:
                    return "teyzesi"
                elif level == 3:
                    return "(anne tarafından) büyük teyzesi"
                elif level == 4:
                    return "(anne tarafından) büyük büyük teyzesi"
            else:
                return "(anne tarafından) %s teyzesi" % _level_name[level - 1]
        else:
            if level == 2:
                return "teyzesi/halası"
            else:
                return "%s teyzesi/halası" % _level_name[level - 1]

    def get_uncle_aunt_unknown(self, level, reltocommon, inlaw=""):
        """
        Return uncle/aunt generation name when gender unknown
        """
        if reltocommon[0] == self.REL_FATHER:
            # from Father
            if level == 2:
                return "babasının kardeşi"
            elif level == 3:
                return "dedesi veya babaannesinin kardeşi"
            else:
                return (
                    "(baba tarafından) %s üstsoyunun kardeşi" % _level_name[level - 1]
                )
        elif reltocommon[0] == self.REL_MOTHER:
            # from Mother
            if level == 2:
                return "annesinin kardeşi"
            elif level == 3:
                return "dedesinin ve anneannesinin kardeşi"
            else:
                return (
                    "(anne tarafından) %s üstsoyunun kardeşi" % _level_name[level - 1]
                )
        else:
            if level == 2:
                return "ebeveynlerinin kardeşi"
            else:
                return "%s üstsoyunun kardeşi" % _level_name[level - 1]

    def get_nephew(self, level, reltocommon, inlaw=""):
        """
        Return nephew generation name
        """
        if inlaw == "":
            t_inlaw = ""
        else:
            t_inlaw = "kayın "

        if level == 2:
            if reltocommon[0] == self.REL_FATHER:
                # Brother's son
                return t_inlaw + "(erkek kardeşinden) yeğeni"
            elif reltocommon[0] == self.REL_MOTHER:
                # Sister's son
                return t_inlaw + "(kız kardeşinden) yeğeni"
            else:
                return t_inlaw + "(kardeşinden) yeğeni"
        elif level >= 2 and level <= 6:
            if reltocommon[level - 2] == self.REL_FATHER:
                # Brother's son generation
                if reltocommon[level - 3] == self.REL_FATHER:
                    # Son of brother son
                    return "erkek kardeşinin oğlunun " + self.get_son(
                        (level - 2), inlaw
                    )
                elif reltocommon[level - 3] == self.REL_MOTHER:
                    # Son of brother daughter
                    return "erkek kardeşinin kızının " + self.get_son(
                        (level - 2), inlaw
                    )
                else:
                    # Currently not used
                    return "erkek kardeşinin çocuğunun " + self.get_son(
                        (level - 2), inlaw
                    )
            elif reltocommon[level - 2] == self.REL_MOTHER:
                # Sister's son generation
                if reltocommon[level - 3] == self.REL_FATHER:
                    # Son of sister son
                    return "kız kardeşinin oğlunun " + self.get_son((level - 2), inlaw)
                elif reltocommon[level - 3] == self.REL_MOTHER:
                    # Son of sister daughter
                    return "kız kardeşinin kızının " + self.get_son((level - 2), inlaw)
                else:
                    # Currently not used
                    return "kız kardeşinin çocuğunun " + self.get_son(
                        (level - 2), inlaw
                    )
            else:
                return "kardeşinin çocuğunun " + self.get_son((level - 2), inlaw)
        else:
            if reltocommon[level - 2] == self.REL_FATHER:
                # Brother's son generation
                return t_inlaw + "erkek kardeşinin oğlunun %d kuşak torunu" % (level)
            elif reltocommon[level - 2] == self.REL_MOTHER:
                # Sister's son generation
                return t_inlaw + "kız kardeşinin oğlunun %d kuşak torunu" % (level)
            else:
                return t_inlaw + "kardeşinin oğlunun %d kuşak torunu" % (level)

    def get_niece(self, level, reltocommon, inlaw=""):
        """
        Return niece generation name
        """
        if inlaw == "":
            t_inlaw = ""
        else:
            t_inlaw = "kayın "

        if level == 2:
            if reltocommon[0] == self.REL_FATHER:
                # Brother's daughter
                return t_inlaw + "(erkek kardeşinden) yeğeni"
            elif reltocommon[0] == self.REL_MOTHER:
                # Sister's daughter
                return t_inlaw + "(kız kardeşinden) yeğeni"
            else:
                return t_inlaw + "(kardeşinden) yeğeni"
        elif level >= 2 and level <= 6:
            if reltocommon[level - 2] == self.REL_FATHER:
                # Brother's daughter generation
                if reltocommon[level - 3] == self.REL_FATHER:
                    # daughter of brother son
                    return "erkek kardeşinin oğlunun " + self.get_daughter(
                        (level - 2), inlaw
                    )
                elif reltocommon[level - 3] == self.REL_MOTHER:
                    # daughter of brother daughter
                    return "erkek kardeşinin kızının " + self.get_daughter(
                        (level - 2), inlaw
                    )
                else:
                    # Currently not used
                    return "erkek kardeşinin çocuğunun " + self.get_daughter(
                        (level - 2), inlaw
                    )
            elif reltocommon[level - 2] == self.REL_MOTHER:
                # Sister's daughter generation
                if reltocommon[level - 3] == self.REL_FATHER:
                    # daughter of sister son
                    return "kız kardeşinin oğlunun " + self.get_daughter(
                        (level - 2), inlaw
                    )
                elif reltocommon[level - 3] == self.REL_MOTHER:
                    # daughter of sister daughter
                    return "kız kardeşinin kızının " + self.get_daughter(
                        (level - 2), inlaw
                    )
                else:
                    # Currently not used
                    return "kız kardeşinin çocuğunun " + self.get_daughter(
                        (level - 2), inlaw
                    )
            else:
                return "kardeşinin çocuğunun " + self.get_daughter(level, inlaw)
        else:
            if reltocommon[level - 2] == self.REL_FATHER:
                # Brother's daughter generation
                return t_inlaw + "erkek kardeşinin kızının %d kuşak torunu" % (level)
            elif reltocommon[level - 2] == self.REL_MOTHER:
                # Sister's daughter generation
                return t_inlaw + "kız kardeşinin kızının %d kuşak torunu" % (level)
            else:
                return t_inlaw + "kardeşinin kızının %d kuşak torunu" % (level)

    def get_nephew_niece_unknown(self, level, reltocommon, inlaw=""):
        """
        Return nephew/niece generation name when gender unknown
        """
        if inlaw == "":
            t_inlaw = ""
        else:
            t_inlaw = "kayın "

        if reltocommon[level - 2] == self.REL_FATHER:
            # Brother's descendant
            return t_inlaw + "erkek kardeşinin " + _child_level[level - 1]
        elif reltocommon[level - 2] == self.REL_MOTHER:
            # Sister's descendant
            return t_inlaw + "kız kardeşinin " + _child_level[level - 1]
        else:
            return t_inlaw + "kardeşinin " + _child_level[level - 1]

    def get_level(self, level, gender):
        """
        Return level name depend of gender
        """
        if gender == Person.MALE:
            if level < len(_level_name_male):
                return _level_name_male[level]
            else:
                return "%d" % level
        elif gender == Person.FEMALE:
            if level < len(_level_name_female):
                return _level_name_female[level]
            else:
                return "%d" % level
        else:
            if level < len(_level_name):
                return _level_name[level]
            else:
                return "%d" % level

    def get_cousin_level(self, level, gender):
        """
        Return level name depend of gender
        """
        if gender == Person.MALE:
            if level < len(_level_name_male):
                return _level_name_male[level - 1]
            else:
                return "%d" % level
        elif gender == Person.FEMALE:
            if level < len(_level_name_female):
                return _level_name_female[level - 1]
            else:
                return "%d" % level
        else:
            if level < len(_level_name):
                return _level_name[level - 1]
            else:
                return "%d" % level

    def get_single_relationship_string(
        self,
        Ga,
        Gb,
        gender_a,
        gender_b,
        reltocommon_a,
        reltocommon_b,
        only_birth=True,
        in_law_a=False,
        in_law_b=False,
    ):
        """
        Provide a string that describes the relationsip between a person, and
        another person. E.g. "grandparent" or "child".

        Ga: The number of generations between the main person and the
            common ancestor.

        Gb: The number of generations between the other person and the
            common ancestor.

        gender_a : gender of person a

        gender_b : gender of person b

        reltocommon_a : relation path to common ancestor or common
                        Family for person a.
                        Note that length = Ga
        reltocommon_b : relation path to common ancestor or common
                        Family for person b.
                        Note that length = Gb

        in_law_a : True if path to common ancestors is via the partner
                          of person a

        in_law_b : True if path to common ancestors is via the partner
                          of person b

        only_birth : True if relation between a and b is by birth only
                            False otherwise

        more see in relationship.py get_single_relationship_string()
        """

        if only_birth:
            step = ""
        else:
            step = self.STEP

        if in_law_a or in_law_b:
            inlaw = self.INLAW
        else:
            inlaw = ""

        t_inlaw = ""

        # b is the same person as a
        if Ga == Gb == 0:
            rel_str = "aynı kişi"

        elif Ga == 0:
            # b is son/descendant of a

            if gender_b == Person.MALE:
                if inlaw and Gb == 1 and not step:
                    # Curretly not used.
                    rel_str = "damadı"
                else:
                    rel_str = self.get_son(Gb, inlaw)

            elif gender_b == Person.FEMALE:
                if inlaw and Gb == 1 and not step:
                    # Curretly not used.
                    rel_str = "gelini"
                else:
                    rel_str = self.get_daughter(Gb, inlaw)

            else:
                rel_str = self.get_child_unknown(Gb, inlaw)

        elif Gb == 0:
            # b is parent/grand parent of a

            if gender_b == Person.MALE:
                if inlaw and Gb == 1 and not step:
                    # Currently nod used.
                    rel_str = "kaynatası"
                else:
                    rel_str = self.get_father(Ga, reltocommon_a, inlaw)

            elif gender_b == Person.FEMALE:
                if inlaw and Gb == 1 and not step:
                    # Currently nod used.
                    rel_str = "kaynanası"
                else:
                    rel_str = self.get_mother(Ga, reltocommon_a, inlaw)

            else:
                rel_str = self.get_parent_unknown(Ga, reltocommon_a, inlaw)

        elif Ga == Gb == 1:
            # Family, brother/sister
            # not used, leave it just in case
            # because should be handled
            # by get_sibling_relationship_string
            # that called in parent RelationshipCalculator
            # see Relationship.py

            if gender_b == Person.MALE:
                if inlaw and not step:
                    rel_str = "erkek kardeşi"
                else:
                    # Currently nod used.
                    rel_str = "üvey erkek kardeşi"

            elif gender_b == Person.FEMALE:
                if inlaw and not step:
                    rel_str = "kız kardeşi"
                else:
                    # Currently nod used.
                    rel_str = "üvey kız kardeşi"
            else:
                rel_str = "kardeşi"

        elif Gb == 1 and Ga > 1:
            # b is aunt/uncle of a

            if gender_b == Person.MALE:
                rel_str = self.get_uncle(Ga, reltocommon_a, inlaw)
            elif gender_b == Person.FEMALE:
                rel_str = self.get_aunt(Ga, reltocommon_a, inlaw)
            else:
                rel_str = self.get_uncle_aunt_unknown(Ga, reltocommon_a, inlaw)

        elif Ga == 1 and Gb > 1:
            # b is niece/nephew of a

            if gender_b == Person.MALE:
                rel_str = self.get_nephew(Gb, reltocommon_b, inlaw)
            elif gender_b == Person.FEMALE:
                rel_str = self.get_niece(Gb, reltocommon_b, inlaw)
            else:
                rel_str = self.get_nephew_niece_unknown(Gb, reltocommon_b, inlaw)

        elif Ga > 1 and Gb > 1:
            # b is (cousin/far aunt/far  uncle/far niece/far nephew) of a

            if Ga > Gb:
                # b is far aunt/far  uncle of a
                level = Ga - Gb + 1
                if level >= 3:  # it is right or???
                    level_name = self.get_level(Gb + 1, gender_b)
                else:
                    level_name = self.get_level(Gb, gender_b)

                if gender_b == Person.MALE:
                    # b is far  uncle
                    if inlaw != "":
                        # Currently not used.
                        t_inlaw = " "
                    if level < len(_parent_level):
                        rel_str = t_inlaw + "%s kuzeni(kuşağı)" % (_parent_level[level])
                    else:
                        rel_str = t_inlaw + "%s üstsoyunun kuzeni(kuşağı)" % (
                            _level_name[level - 1]
                        )
                elif gender_b == Person.FEMALE:
                    # b is far aunt
                    if inlaw != "":
                        # Currently not used.
                        t_inlaw = " "
                    if level < len(_aunt_level):
                        rel_str = t_inlaw + "%s kuzeni(kuşağı)" % (_parent_level[level])
                    else:
                        rel_str = t_inlaw + "%s üstsoyunun kuzeni(kuşağı)" % (
                            _level_name[level - 1]
                        )
                else:
                    if inlaw != "":
                        # Currently not used.
                        t_inlaw = " "
                    if level == 2:
                        rel_str = t_inlaw + "%s kuzeni(kuşağı)" % (_parent_level[level])
                    else:
                        rel_str = t_inlaw + "%s üstsoyunun kuzeni(kuşağı)" % (
                            _level_name[level - 1]
                        )

            elif Ga < Gb:
                # b is far niece/far nephew of a
                level_name = self.get_level(Ga, gender_b)
                level = Gb - Ga + 1

                if gender_b == Person.MALE:
                    # b is far nephew
                    if level == 2:
                        rel_str = "%s yeğeni" % level_name
                    else:
                        rel_str = "%s yeğeninin %s" % (
                            level_name,
                            _son_level[level - 2],
                        )
                elif gender_b == Person.FEMALE:
                    # b is far niece
                    if level == 2:
                        rel_str = "%s yeğeni" % level_name
                    else:
                        rel_str = "%s yeğeninin %s" % (
                            level_name,
                            _daughter_level[level - 2],
                        )
                else:
                    if level == 2:
                        rel_str = "%s yeğeni" % level_name
                    else:
                        rel_str = "%s yeğeninin %s" % (
                            level_name,
                            _child_level[level - 2],
                        )

            else:  # Gb == Ga
                # b is cousin of a
                level_name = self.get_cousin_level(Ga, gender_b)
                if gender_b == Person.MALE:
                    if inlaw != "":
                        # Currently not used.
                        t_inlaw = " "
                    rel_str = t_inlaw + "%s kuzeni" % level_name
                elif gender_b == Person.FEMALE:
                    if inlaw != "":
                        # Currently not used.
                        t_inlaw = " "
                    rel_str = t_inlaw + "%s kuzeni" % level_name
                else:
                    if inlaw != "":
                        # Currently not used.
                        t_inlaw = " "
                    rel_str = t_inlaw + "%s kuzeni" % level_name

        else:
            # A program should never goes there, but...
            rel_str = "sonsuz ilişki derecesi"

        return rel_str

    def get_plural_relationship_string(
        self,
        Ga,
        Gb,
        reltocommon_a="",
        reltocommon_b="",
        only_birth=True,
        in_law_a=False,
        in_law_b=False,
    ):
        """
        Generate a text with information, how far away is a group of persons
        from a main person
        """
        if Ga == Gb == 0:
            return "aynı kişi"
        if 0 == Ga:
            # These are descendants
            if 1 == Gb:
                return "çocukları"
            if 2 == Gb:
                return "torunları"
            if 3 == Gb:
                return "torunlarının çocukları"
            if 4 == Gb:
                return "torunlarının torunları"
            return "torunlarının torunlarından altsoya ait torunları"
        if 0 == Gb:
            # These are parents/grand parents
            if 1 == Ga:
                return "ebeveynleri"
            if 2 == Ga:
                return "dedeleri ve nineleri"
            if 3 == Ga:
                return "büyük dedeleri ve haminneleri"
            if 4 == Ga:
                return "büyük büyük dedeleri ve nineleri"
            return "büyük büyük dedeleri ve ninelerinden üstsoya ait ataları"
        if 1 == Ga == Gb:
            return "kardeşleri"
        if 1 == Gb and Ga > 1:
            return "ebeveynlerinin kardeşleri"
        if 1 < Gb and 1 == Ga:
            return "yeğenleri"
        if 1 < Ga and 1 < Gb:
            return "uzak akrabaları"
        return "arasındaki ilişki belirsiz"

    def get_sibling_relationship_string(
        self, sib_type, gender_a, gender_b, in_law_a=False, in_law_b=False
    ):
        if in_law_a or in_law_b:
            inlaw = self.INLAW
        else:
            inlaw = ""

        if sib_type == self.NORM_SIB:
            if not inlaw:
                if gender_b == Person.MALE:
                    rel_str = "erkek kardeşi"
                elif gender_b == Person.FEMALE:
                    rel_str = "kız kardeşi"
                else:
                    rel_str = "kardeşi"
            else:
                if gender_b == Person.MALE:
                    rel_str = "erkek kardeşi"
                elif gender_b == Person.FEMALE:
                    rel_str = "kız kardeşi"
                else:
                    rel_str = "kardeşi"
        elif sib_type == self.UNKNOWN_SIB:
            if not inlaw:
                if gender_b == Person.MALE:
                    rel_str = "erkek kardeşi"
                elif gender_b == Person.FEMALE:
                    rel_str = "kız kardeşi"
                else:
                    rel_str = "kardeşi"
            else:
                if gender_b == Person.MALE:
                    rel_str = "erkek kardeşi"
                elif gender_b == Person.FEMALE:
                    rel_str = "kız kardeşi"
                else:
                    rel_str = "kardeşi"
        else:
            rel_str = "arasındaki ilişki belirsiz"
        return rel_str


if __name__ == "__main__":
    # Test function. Call it as follows from the command line (so as to find
    #        imported modules):
    #    export PYTHONPATH=/path/to/gramps/src
    #    python src/plugins/rel/rel_tr.py

    """
    TRANSLATORS, copy this if statement at the bottom of your
    rel_xx.py module, and test your work with:
    python src/plugins/rel/rel_xx.py
    """
    from gramps.gen.relationship import test

    RC = RelationshipCalculator()
    test(RC, True)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2016 Tom Samstag
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Unittest that tests media-specific filter rules
"""
import unittest
import os

from ....db.utils import import_as_dict
from ....filters import GenericFilterFactory
from ....const import DATA_DIR
from ....user import User

from ..media import (
    AllMedia,
    HasIdOf,
    RegExpIdOf,
    HasCitation,
    HasNoteRegexp,
    HasNoteMatchingSubstringOf,
    HasReferenceCountOf,
    HasSourceCount,
    HasSourceOf,
    MediaPrivate,
    MatchesSourceConfidence,
    HasMedia,
    HasAttribute,
    ChangedSince,
    HasTag,
)

TEST_DIR = os.path.abspath(os.path.join(DATA_DIR, "tests"))
EXAMPLE = os.path.join(TEST_DIR, "example.gramps")
GenericMediaFilter = GenericFilterFactory("Media")


class BaseTest(unittest.TestCase):
    """
    Media rule tests.
    """

    @classmethod
    def setUpClass(cls):
        """
        Import example database.
        """
        cls.db = import_as_dict(EXAMPLE, User())

    def filter_with_rule(self, rule):
        """
        Apply a filter with the given rule.
        """
        filter_ = GenericMediaFilter()
        filter_.add_rule(rule)
        results = filter_.apply(self.db)
        return set(results)

    def test_allmedia(self):
        """
        Test AllMedia rule.
        """
        rule = AllMedia([])
        self.assertEqual(
            len(self.filter_with_rule(rule)), self.db.get_number_of_media()
        )

    def test_hasidof(self):
        """
        Test HasIdOf rule.
        """
        rule = HasIdOf(["O0000"])
        self.assertEqual(self.filter_with_rule(rule), set(["b39fe1cfc1305ac4a21"]))

    def test_regexpidof(self):
        """
        Test RegExpIdOf rule.
        """
        rule = RegExpIdOf(["O000."], use_regex=True)
        self.assertEqual(
            self.filter_with_rule(rule),
            set(
                [
                    "F0QIGQFT275JFJ75E8",
                    "78V2GQX2FKNSYQ3OHE",
                    "b39fe1cfc1305ac4a21",
                    "F8JYGQFL2PKLSYH79X",
                    "B1AUFQV7H8R9NR4SZM",
                ]
            ),
        )

    def test_hascitation(self):
        """
        Test HasCitation rule.
        """
        rule = HasCitation(["page 21", "", ""])
        self.assertEqual(self.filter_with_rule(rule), set(["B1AUFQV7H8R9NR4SZM"]))

    def test_hasnoteregexp(self):
        """
        Test HasNoteRegexp rule.
        """
        rule = HasNoteRegexp(["."], use_regex=True)
        self.assertEqual(self.filter_with_rule(rule), set([]))

    def test_hasnotematchingsubstringof(self):
        """
        Test HasNoteMatchingSubstringOf rule.
        """
        rule = HasNoteMatchingSubstringOf(["some text"])
        self.assertEqual(self.filter_with_rule(rule), set([]))

    def test_hasreferencecountof(self):
        """
        Test HasReferenceCountOf rule.
        """
        rule = HasReferenceCountOf(["greater than", "1"])
        self.assertEqual(
            self.filter_with_rule(rule),
            set(["238CGQ939HG18SS5MG", "b39fe1cfc1305ac4a21", "Y3ARGQWE088EQRTTDH"]),
        )

    def test_hassourcecount(self):
        """
        Test HasSourceCount rule.
        """
        rule = HasSourceCount(["1", "equal to"])
        self.assertEqual(self.filter_with_rule(rule), set(["B1AUFQV7H8R9NR4SZM"]))

    def test_hassourceof(self):
        """
        Test HasSourceOf rule.
        """
        rule = HasSourceOf(["S0001"])
        self.assertEqual(self.filter_with_rule(rule), set(["B1AUFQV7H8R9NR4SZM"]))

    def test_mediaprivate(self):
        """
        Test MediaPrivate rule.
        """
        rule = MediaPrivate([])
        self.assertEqual(self.filter_with_rule(rule), set([]))

    def test_matchessourceconfidence(self):
        """
        Test MatchesSourceConfidence rule.
        """
        rule = MatchesSourceConfidence(["2"])
        self.assertEqual(self.filter_with_rule(rule), set(["B1AUFQV7H8R9NR4SZM"]))

    def test_hasmedia(self):
        """
        Test HasMedia rule.
        """
        rule = HasMedia(["mannschaft", "image/jpeg", ".jpg", "1897"])
        self.assertEqual(self.filter_with_rule(rule), set(["238CGQ939HG18SS5MG"]))

    def test_hasattribute(self):
        """
        Test HasAttribute rule.
        """
        rule = HasAttribute(["Description", ""])
        self.assertEqual(self.filter_with_rule(rule), set(["B1AUFQV7H8R9NR4SZM"]))

    def test_changedsince(self):
        """
        Test ChangedSince rule.
        """
        rule = ChangedSince(["2010-01-01", "2016-01-01"])
        self.assertEqual(
            self.filter_with_rule(rule),
            set(
                [
                    "b39fe1cfc1305ac4a21",
                    "B1AUFQV7H8R9NR4SZM",
                    "238CGQ939HG18SS5MG",
                    "F0QIGQFT275JFJ75E8",
                ]
            ),
        )

    def test_hastag(self):
        """
        Test HasTag rule.
        """
        rule = HasTag(["ToDo"])
        self.assertEqual(self.filter_with_rule(rule), set(["238CGQ939HG18SS5MG"]))


if __name__ == "__main__":
    unittest.main()



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2002-2006  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# gen.filters.rules/Place/_HasNoLatOrLon.py


# -------------------------------------------------------------------------
#
# Standard Python modules
#
# -------------------------------------------------------------------------
from ....const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from .. import Rule


# -------------------------------------------------------------------------
#
# Typing modules
#
# -------------------------------------------------------------------------
from ....lib import Place
from ....db import Database


# -------------------------------------------------------------------------
#
# HasNoLatOrLon
#
# -------------------------------------------------------------------------
class HasNoLatOrLon(Rule):
    """Rule that checks if Latitude or Longitude are not given"""

    labels = []
    name = _("Places with no latitude or longitude given")
    description = _("Matches places with empty latitude or longitude")
    category = _("Position filters")

    def apply_to_one(self, db: Database, place: Place) -> bool:
        if place.lat.strip() and place.long.strip():
            return False
        return True



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2004-2006  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Class handling language-specific selection for date parser and displayer.
"""

# -------------------------------------------------------------------------
#
# Python modules
#
# -------------------------------------------------------------------------
import time

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from ..const import GRAMPS_LOCALE as glocale
from ..lib.date import Date
from . import LANG_TO_DISPLAY, LANG, parser, displayer


# --------------------------------------------------------------
#
# Convenience functions
#
# --------------------------------------------------------------
def get_date_formats(flocale=glocale):
    """
    Return the list of supported formats for date parsers and displayers.
    The UI language formats will be used unless another locale is fed in.

    :param flocale: allow deferred translation of date formats
    :type flocale: a :class:`.GrampsLocale` instance
    """
    # trans_text is a defined keyword (see po/update_po.py, po/genpot.sh)
    trans_text = flocale.translation.sgettext
    try:
        return tuple(
            trans_text(fmt) for fmt in LANG_TO_DISPLAY[flocale.lang](0).formats
        )
    except:
        return tuple(trans_text(fmt) for fmt in LANG_TO_DISPLAY["C"](0).formats)


def set_format(value):
    try:
        displayer.set_format(value)
    except:
        pass


def set_date(date_base, text):
    """
    Set the date of the :class:`.DateBase` instance.

    The date is parsed into a :class:`.Date` instance.

    :param date_base: The :class:`.DateBase` instance to set the date to.
    :type date_base: :class:`.DateBase`
    :param text: The text to use for the text string in date
    :type text: str
    """
    parser.set_date(date_base.get_date_object(), text)


def get_date(date_base):
    """
    Return a string representation of the date of the :class:`.DateBase`
    instance.

    This representation is based off the default date display format
    determined by the locale's :class:`.DateDisplay` instance.

    :return: Returns a string representing the :class:`.DateBase` date
    :rtype: str
    """
    return displayer.display(date_base.get_date_object())


def get_date_valid(date_base):
    date_obj = date_base.get_date_object()
    return date_obj.get_valid()


def format_time(secs):
    """
    Format a time in seconds as a date in the preferred date format and a
    24 hour time as hh:mm:ss.
    """
    t = time.localtime(secs)
    d = Date(t.tm_year, t.tm_mon, t.tm_mday)
    return displayer.display(d) + time.strftime(" %X", t)



#! /usr/bin/env python3
"""Test program for import modules"""
#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (c) 2016 Gramps Development Team
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
#
# **********************************************
# This module is used to compare dbs from imported (i.e. GEDCOM) or old versions
# of the Gramps db directories with correct gramps XML files.
# This will highlight changes to the db caused by changes in the code.
# **********************************************

import unittest
import os
import sys
import re
import locale
from time import localtime, strptime
from unittest.mock import patch
import zipfile
import shutil

# import logging

from gramps.gen.utils.config import config

config.set("preferences.date-format", 0)
from gramps.gen.db.utils import import_as_dict
from gramps.gen.merge.diff import diff_dbs
from gramps.gen.lib.json_utils import object_to_dict
from gramps.gen.simple import SimpleAccess
from gramps.gen.utils.id import set_det_id
from gramps.gen.user import User
from gramps.gen.const import USER_DATA, DATA_DIR
from gramps.test.test_util import capture
from gramps.plugins.export.exportxml import XmlWriter
from gramps.gen.db.dbconst import DBBACKEND
from gramps.gen.db.utils import make_database
from gramps.gen.errors import DbError
from gramps.gen.db.exceptions import (
    DbUpgradeRequiredError,
    DbVersionError,
    DbPythonError,
    DbSupportedError,
    DbConnectionError,
)

# logger = logging.getLogger(__name__)

# the following defines where to find the test import and result files
TEST_DIR = os.path.abspath(os.path.join(DATA_DIR, "tests"))
# the following defines where to find test error diffs and import result XML files
TEMP_DIR = os.path.join(USER_DATA, "temp")
if not os.path.isdir(TEMP_DIR):
    os.makedirs(TEMP_DIR)

# ------------------------------------------------------------------
#  Local Functions
# ------------------------------------------------------------------


def mock_time(*args):
    """
    Mock up a dummy to replace the varying 'time string results'
    """
    return 946101600.0


def mock_localtime(*args):
    """
    Mock up a dummy to replace the varying 'time string results'
    """
    return strptime("25 Dec 1999", "%d %b %Y")


# These tests assume a US date and time format.
# If the locale is not available on the
# build host, skip these tests.
en_US_locale_available = False
try:
    locale.setlocale(locale.LC_ALL, "en_US.utf8")
    en_US_locale_available = True
except locale.Error:  # seems to fail on Windows system for some reason
    try:
        locale.setlocale(locale.LC_ALL, "English_United States")
        en_US_locale_available = True
    except locale.Error:
        pass


@unittest.skipUnless(en_US_locale_available, "en_US locale is not avaiable")
class TestImports(unittest.TestCase):
    """The test class cases will be dynamically created at import time from
    files to be tested.  The following defs are used by the test cases
    """

    def prepare_result(self, diffs, added, missing):
        """Looks through the diffs, added, and missing items and begins
        reporting process.  Returns True if there were significant errors.
        """
        # pylint: disable=E1101
        # pylint does not like dynamically created methods
        deltas = False
        if diffs:
            for diff in diffs:
                obj_type, item1, item2 = diff
                msg = self._report_diff(
                    obj_type, object_to_dict(item1), object_to_dict(item2)
                )
                if msg != "":
                    if hasattr(item1, "gramps_id"):
                        self.msg += "%s: %s  handle=%s\n" % (
                            obj_type,
                            item1.gramps_id,
                            getattr(item1, "handle"),
                        )
                    else:
                        self.msg += "%s: %s\n" % (obj_type, item1.get_name())
                    self.msg += msg
                    deltas = True
        if missing:
            deltas = True
            sac = SimpleAccess(self.database1)
            for pair in missing:
                obj_type, item = pair
                self.msg += "Missing %s: %s\n" % (obj_type, sac.describe(item))
        if added:
            deltas = True
            sac = SimpleAccess(self.database2)
            for pair in added:
                obj_type, item = pair
                self.msg += "Added %s: %s\n" % (obj_type, sac.describe(item))
        return deltas

    def _report_diff(self, path, struct1, struct2):
        """
        Compare two struct objects and report differences.
        """
        msg = ""
        if struct1 == struct2:
            return ""
        elif isinstance(struct1, (list, tuple)) or isinstance(struct2, (list, tuple)):
            len1 = len(struct1) if isinstance(struct1, (list, tuple)) else 0
            len2 = len(struct2) if isinstance(struct2, (list, tuple)) else 0
            for pos in range(max(len1, len2)):
                value1 = struct1[pos] if pos < len1 else None
                value2 = struct2[pos] if pos < len2 else None
                msg += self._report_diff(path + ("[%d]" % pos), value1, value2)
        elif isinstance(struct1, dict) or isinstance(struct2, dict):
            keys = struct1.keys() if isinstance(struct1, dict) else struct2.keys()
            for key in keys:
                value1 = struct1[key] if struct1 is not None else None
                value2 = struct2[key] if struct2 is not None else None
                if key == "dict":  # a raw dict, not a struct
                    msg += _report_details(path, value1, value2)
                else:
                    msg += self._report_diff(path + "." + key, value1, value2)
        else:
            msg += _report_details(path, struct1, struct2)
        return msg


def _report_details(path, diff1, diff2):
    """Checks if a detail is significant, needs adjusting for xml filter
    effects, and returns a string describing the specific difference.
    """
    if isinstance(diff1, bool):
        desc1 = repr(diff1)
    else:
        desc1 = str(diff1) if diff1 else ""
    if isinstance(diff2, bool):
        desc2 = repr(diff2)
    else:
        desc2 = str(diff2) if diff2 else ""
    d1t = type(diff1)
    d2t = type(diff2)
    # the xml exporter edits the data base by stripping spaces, so
    # we have to ignore these differences
    if d1t == str:
        diff1 = diff1.strip()
    if d2t == str:
        diff2 = diff2.strip()
    d1l = len(diff1) if d1t == str else ""
    d2l = len(diff2) if d2t == str else ""
    # 'change' date is not significant for comparison
    if path.endswith(".change"):
        return ""
    # the xml exporter edits the data base by converting media path
    # to unix '/' conventions, so we have to ignore these differences
    if path == "Media.path":
        diff1 = diff1.replace("\\", "/")
    if diff1 != diff2:
        msg = "  Diff on: %s\n    %s%s: %s\n    %s%s: %s\n" % (
            _format_struct_path(path),
            d1t,
            d1l,
            desc1,
            d2t,
            d2l,
            desc2,
        )
        return msg
    return ""


def _format_struct_path(path):
    """prepares a 'path' string for the report out of the structure."""
    retval = ""
    parts = path.split(".")
    for part in parts:
        if retval:
            retval += ", "
        if "[" in part and "]" in part:
            part, index = re.match(r"(.*)\[(\d*)\]", part).groups()
            retval += "%s #%s" % (part.replace("_", " "), int(index) + 1)
        else:
            retval += part
    return retval


def db_load(zipfn, self):
    """load a legacy db (typically bsddb, but sqlite allowed)
    Note: zipfn is a location for zip of the the db directory.
    To make this code easier, the db dir must be named with same root name as the
    zip file.  For example the "imp_413.zip" must contain a single "imp_413" dir
    (not the usual 8 hex character name).
    """
    (tstfile, _ext) = os.path.splitext(zipfn)
    tstfile = os.path.join(TEMP_DIR, os.path.basename(tstfile))
    shutil.rmtree(tstfile, ignore_errors=True)
    with zipfile.ZipFile(zipfn, "r") as myzip:
        myzip.extractall(path=TEMP_DIR)
    force_schema_upgrade = True
    try:
        while True:
            dbid_path = os.path.join(tstfile, DBBACKEND)
            if os.path.isfile(dbid_path):
                with open(dbid_path) as fp:
                    dbid = fp.read().strip()
            else:
                dbid = "bsddb"

            db = make_database(dbid)
            db.disable_signals()

            try:
                db.load(
                    tstfile,
                    callback=None,
                    mode="w",
                    force_schema_upgrade=force_schema_upgrade,
                )
                break
            except (DbSupportedError, DbUpgradeRequiredError) as msg:
                continue
        return db
    # Get here is there is an exception the while loop does not handle
    except (
        DbVersionError,
        DbPythonError,
        DbConnectionError,
        OSError,
        DbError,
        Exception,
    ) as msg:
        msg = str(msg)
        self.assertTrue(False, msg=f"Cannot open database {tstfile} {msg}")
    return


def make_tst_function(tstfile, file_name):
    """This is here to support the dynamic function creation.  This creates
    the test function (a method, to be precise).
    """

    @patch("gramps.plugins.db.dbapi.dbapi.time")
    @patch("gramps.gen.utils.unknown.localtime")
    @patch("gramps.gen.utils.unknown.time")
    @patch("time.localtime")
    def tst(self, mockptime, mocktime, mockltime, mockdtime):
        """This compares the import file with the expected result '.gramps'
        file.
        """
        mockptime.side_effect = mock_localtime
        mocktime.side_effect = mock_time
        mockltime.side_effect = mock_localtime
        mockdtime.side_effect = mock_time
        fn1 = os.path.join(TEST_DIR, tstfile)
        fn2 = os.path.join(TEST_DIR, (file_name + ".gramps"))
        fres = os.path.join(TEMP_DIR, (file_name + ".difs"))
        fout = os.path.join(TEMP_DIR, (file_name + ".gramps"))
        if "_dfs" in tstfile:
            config.set("preferences.default-source", True)
            config.set("preferences.tag-on-import-format", "Imported")
            config.set("preferences.tag-on-import", True)
            skp_imp_adds = False
        else:
            skp_imp_adds = True
            config.set("preferences.default-source", False)
            config.set("preferences.tag-on-import", False)
        try:
            os.remove(fres)
            os.remove(fout)
        except OSError:
            pass
        # logger.info("\n**** %s ****", tstfile)
        set_det_id(True)
        with capture(None) as output:
            self.user = User()
            if ".zip" in fn1:  # must be a db, not an importable file
                self.database1 = db_load(fn1, self)
            else:
                self.database1 = import_as_dict(
                    fn1, self.user, skp_imp_adds=skp_imp_adds
                )
            set_det_id(True)
            self.database2 = import_as_dict(fn2, self.user)
            msg = (
                "\n****Captured Output****\n"
                + str(output[0])
                + "\n****Captured Err****\n"
                + str(output[1])
                + "\n****End Capture Err****\n"
            )
        self.assertIsNotNone(self.database1, f"Unable to import file: {fn1} {msg}")
        self.assertIsNotNone(
            self.database2, f"Unable to import expected result file: {fn2} {msg}"
        )
        if self.database2 is None or self.database1 is None:
            return
        diffs, added, missing = diff_dbs(self.database1, self.database2, self.user)
        self.msg = "Mismatch on file: %s\n" % tstfile
        deltas = self.prepare_result(diffs, added, missing)
        # We save a copy of any issues in the users Gramps TEMP_DIR in a file
        # with a '.difs' extension, as well as usual unittest report
        # Also we save the .gramps file from the import so user can perform
        # Diff himself for better context.
        if deltas:
            writer = XmlWriter(self.database1, self.user, strip_photos=0, compress=0)
            writer.write(fout)
            hres = open(fres, mode="w", encoding="utf-8", errors="replace")
            hres.write(self.msg)
            hres.close()
            # let's see if we have any allowed exception file
            fdif = os.path.join(TEST_DIR, (file_name + ".difs"))
            try:
                hdif = open(fdif)
                msg = hdif.read()
                hdif.close()
            except (FileNotFoundError, IOError):
                msg = ""
            # if exception file matches exactly, we are done.
            if self.msg != msg:
                self.msg = (
                    "\n****Captured Output****\n"
                    + output[0]
                    + "\n****Captured Err****\n"
                    + output[1]
                    + "\n****End Capture Err****\n"
                    + self.msg
                )
                self.fail(self.msg)
        if ".zip" in tstfile:
            fn1 = self.database1.path
            self.database1.close(update=False)
            shutil.rmtree(fn1, ignore_errors=True)

    return tst


# let's see if we have a single file to run, example;
#    "python test_import.py -i sample.ged"
# This only works for files in normal test directory, so don't add a path
# pylint: disable=invalid-name
_tstfile = ""
if __name__ == "__main__":
    for i, option in enumerate(sys.argv):
        if option == "-i":
            _tstfile = sys.argv[i + 1]
            del sys.argv[i]
            del sys.argv[i]
# The following code dynamically creates the methods for each test file.
# The methods are inserted at load time into the 'TestImports' class
# via the modules' globals, taking advantage that they are a dict.
if _tstfile:  # single file mode
    (fname, ext) = os.path.splitext(os.path.basename(_tstfile))
    test_func = make_tst_function(_tstfile, fname)
    tname = "test_" + _tstfile.replace("-", "_").replace(".", "_")
    test_func.__name__ = tname
    test_func.__doc__ = tname
    setattr(TestImports, tname, test_func)
else:
    for _tstfile in os.listdir(TEST_DIR):
        (fname, ext) = os.path.splitext(os.path.basename(_tstfile))
        if _tstfile != "SAMPLE.DEF" and (
            ext in (".gramps", ".difs", ".bak") or not fname.startswith("imp_")
        ):
            continue
        test_func = make_tst_function(_tstfile, fname)
        tname = "test_" + _tstfile.replace("-", "_").replace(".", "_")
        test_func.__name__ = tname
        test_func.__doc__ = tname
        setattr(TestImports, tname, test_func)
    test_func = None  # keep nosetest from finding last one again

if __name__ == "__main__":
    unittest.main()



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2002-2006  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Rule that checks for an event with a particular value.
"""

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from ...const import GRAMPS_LOCALE as glocale
from ...datehandler import parser
from ...display.place import displayer as place_displayer
from ...lib.eventtype import EventType
from ...utils.db import get_participant_from_event
from . import Rule

_ = glocale.translation.gettext


# -------------------------------------------------------------------------
#
# Typing modules
#
# -------------------------------------------------------------------------
from ...lib import Event
from ...db import Database


# -------------------------------------------------------------------------
#
# HasEventBase
#
# -------------------------------------------------------------------------
class HasEventBase(Rule):
    """
    Rule that checks for an event with a particular value.
    """

    labels = ["Event type:", "Date:", "Place:", "Description:", "Main Participants:"]
    name = "Events matching parameters"
    description = "Matches events with particular parameters"
    category = _("Event filters")
    allow_regex = True

    def __init__(self, arg, use_regex=False, use_case=False):
        super().__init__(arg, use_regex, use_case)
        self.date = None
        self.event_type = None

    def prepare(self, db: Database, user):
        """
        Prepare the rule. Things that should only be done once.
        """
        if self.list[0]:
            self.event_type = EventType()
            self.event_type.set_from_xml_str(self.list[0])
        try:
            if self.list[1]:
                self.date = parser.parse(self.list[1])
        except:
            pass

    def apply_to_one(self, db: Database, event: Event) -> bool:
        """
        Apply the rule. Return True if a match.
        """
        if self.event_type:
            if self.event_type.is_custom() and self.use_regex:
                if self.regex[0].search(str(event.type)) is None:
                    return False
            elif event.type != self.event_type:
                return False

        if not self.match_substring(3, event.description):
            return False

        if self.date:
            if not event.date.match(self.date):
                return False

        if self.list[2]:
            place_id = event.place
            if place_id:
                place = db.get_place_from_handle(place_id)
                place_title = place_displayer.display(db, place)
                if not self.match_substring(2, place_title):
                    return False
            else:
                return False

        if not self.match_substring(
            4, get_participant_from_event(db, event.handle, all_=True)
        ):
            return False

        return True



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2016 Tom Samstag
# Copyright (C) 2025 Doug Blank
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Unittest that tests person-specific filter rules
"""
import unittest
import os

from ...db.utils import import_as_dict
from ...const import DATA_DIR
from ...user import User
from ...lib.person import Person
from ...lib.json_utils import remove_object

from ...proxy import PrivateProxyDb, LivingProxyDb

TEST_DIR = os.path.abspath(os.path.join(DATA_DIR, "tests"))
EXAMPLE = os.path.join(TEST_DIR, "example.gramps")


class PrivateProxyTest(unittest.TestCase):
    """
    Person rule tests.
    """

    @classmethod
    def setUpClass(cls):
        """
        Import example database.
        """
        cls.db = PrivateProxyDb(import_as_dict(EXAMPLE, User()))

    def assertDataEquals(self, data1, data2):
        self.assertIsInstance(data1, dict)
        self.assertIsInstance(data2, dict)
        self.assertEqual(
            remove_object(data1),
            remove_object(data2),
        )

    def test_person_count(self):
        count = self.db.get_number_of_people()
        # One private person:
        self.assertEqual(count, 2127)
        # No proxy:
        count = self.db.basedb.get_number_of_people()
        self.assertEqual(count, 2128)

    def test_private_person_data(self):
        # A private person:
        handle = "0GDKQC54XKSWZKEBWW"
        data = self.db.get_raw_person_data(handle)
        self.assertIs(data, None)
        # But they are visible without proxy:
        data = self.db.basedb.get_raw_person_data(handle)
        self.assertIsNot(data, None)

    def test_not_private_person_data(self):
        # A non-private person:
        handle = "0FWJQCLYEP736P3YZK"
        data = self.db.get_raw_person_data(handle)
        self.assertIsNot(data, None)
        self.assertIsInstance(data, dict)
        self.assertEqual(data.handle, handle)
        # Should be same:
        data_orig = self.db.basedb.get_raw_person_data(handle)
        self.assertDataEquals(data, data_orig)

    def test_private_person(self):
        # A private person:
        handle = "0GDKQC54XKSWZKEBWW"
        person = self.db.get_person_from_handle(handle)
        self.assertIs(person, None)
        # But they are visible without proxy:
        person = self.db.basedb.get_person_from_handle(handle)
        self.assertIsNot(person, None)

    def test_not_private_person(self):
        # A non-private person:
        handle = "0FWJQCLYEP736P3YZK"
        person = self.db.get_person_from_handle(handle)
        self.assertIsNot(person, None)
        self.assertIsInstance(person, Person)
        person_orig = self.db.basedb.get_person_from_handle(handle)
        # Same person:
        self.assertEqual(person.gramps_id, person_orig.gramps_id)

    def test_person_with_private_data(self):
        # A person with private data:
        handle = "GNUJQCL9MD64AM56OH"
        person = self.db.get_person_from_handle(handle)
        self.assertTrue(len(person.attribute_list) == 2)
        # Unfiltered person has three attributes
        person = self.db.basedb.get_person_from_handle(handle)
        self.assertTrue(len(person.attribute_list) == 3)


class LivingProxyTest(unittest.TestCase):
    """
    Person rule tests.
    """

    @classmethod
    def setUpClass(cls):
        """
        Import example database.
        """
        cls.db = LivingProxyDb(
            import_as_dict(EXAMPLE, User()),
            mode=LivingProxyDb.MODE_EXCLUDE_ALL,
            current_year=2006,
            years_after_death=10,
        )

    def test_person_count(self):
        count = self.db.get_number_of_people()
        self.assertEqual(count, 1245)

    def test_live_person_data(self):
        handle = "004KQCGYT27EEPQHK"
        data = self.db.get_raw_person_data(handle)
        self.assertIsInstance(data, dict)
        self.assertEqual(data.primary_name.first_name, "Martha")

    def test_dead_person_data(self):
        handle = "66TJQC6CC7ZWL9YZ64"
        data = self.db.get_raw_person_data(handle)
        self.assertIs(data, None)

    def test_live_person(self):
        handle = "004KQCGYT27EEPQHK"
        person = self.db.get_person_from_handle(handle)
        self.assertIsInstance(person, Person)

    def test_dead_person(self):
        handle = "66TJQC6CC7ZWL9YZ64"
        person = self.db.get_person_from_handle(handle)
        self.assertIs(person, None)


class LivingPrivateProxyTest(unittest.TestCase):
    """
    Person rule tests.
    """

    @classmethod
    def setUpClass(cls):
        """
        Import example database.
        """
        cls.db = LivingProxyDb(
            PrivateProxyDb(import_as_dict(EXAMPLE, User())),
            mode=LivingProxyDb.MODE_EXCLUDE_ALL,
            current_year=2006,
            years_after_death=10,
        )

    def test_person_count(self):
        count = self.db.get_number_of_people()
        self.assertEqual(count, 1244)

    def test_private_person(self):
        # A private person:
        handle = "0GDKQC54XKSWZKEBWW"
        person = self.db.get_person_from_handle(handle)
        self.assertIs(person, None)
        # But they are visible without proxy:
        person = self.db.basedb.get_person_from_handle(handle)
        self.assertIsNot(person, None)



# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2001-2006  Donald N. Allingham
# Copyright (C) 2008       Gary Burton
# Copyright (C) 2011       Tim G L Lyons
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Source View
"""

# -------------------------------------------------------------------------
#
# GTK/Gnome modules
#
# -------------------------------------------------------------------------
from gi.repository import Gtk
import logging

LOG = logging.getLogger(".citation")

# -------------------------------------------------------------------------
#
# gramps modules
#
# -------------------------------------------------------------------------
from gramps.gen.lib import Source
from gramps.gen.config import config
from gramps.gui.views.listview import ListView, TEXT, MARKUP, ICON
from gramps.gui.views.treemodels import SourceModel
from gramps.gui.views.bookmarks import SourceBookmarks
from gramps.gen.errors import WindowActiveError
from gramps.gui.ddtargets import DdTargets
from gramps.gui.dialog import ErrorDialog
from gramps.gui.editors import EditSource
from gramps.gui.filters.sidebar import SourceSidebarFilter
from gramps.gui.merge import MergeSource
from gramps.gen.plug import CATEGORY_QR_SOURCE
from gramps.plugins.lib.libsourceview import LibSourceView

# -------------------------------------------------------------------------
#
# internationalization
#
# -------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.sgettext


# -------------------------------------------------------------------------
#
# SourceView
#
# -------------------------------------------------------------------------
class SourceView(LibSourceView, ListView):
    """sources listview class"""

    COL_TITLE = 0
    COL_ID = 1
    COL_AUTH = 2
    COL_ABBR = 3
    COL_PINFO = 4
    COL_PRIV = 5
    COL_TAGS = 6
    COL_CHAN = 7

    # column definitions
    COLUMNS = [
        (_("Title"), TEXT, None),
        (_("ID"), TEXT, None),
        (_("Author"), TEXT, None),
        (_("Abbreviation"), TEXT, None),
        (_("Publication Information"), TEXT, None),
        (_("Private"), ICON, "gramps-lock"),
        (_("Tags"), TEXT, None),
        (_("Last Changed"), TEXT, None),
    ]
    # default setting with visible columns, order of the col, and their size
    CONFIGSETTINGS = (
        ("columns.visible", [COL_TITLE, COL_ID, COL_AUTH, COL_PINFO]),
        (
            "columns.rank",
            [
                COL_TITLE,
                COL_ID,
                COL_AUTH,
                COL_ABBR,
                COL_PINFO,
                COL_PRIV,
                COL_TAGS,
                COL_CHAN,
            ],
        ),
        ("columns.size", [200, 75, 150, 100, 150, 40, 100, 100]),
    )
    ADD_MSG = _("Add a new source")
    EDIT_MSG = _("Edit the selected source")
    DEL_MSG = _("Delete the selected source")
    MERGE_MSG = _("Merge the selected sources")
    FILTER_TYPE = "Source"
    QR_CATEGORY = CATEGORY_QR_SOURCE

    def __init__(self, pdata, dbstate, uistate, nav_group=0):
        signal_map = {
            "source-add": self.row_add,
            "source-update": self.row_update,
            "source-delete": self.row_delete,
            "source-rebuild": self.object_build,
        }

        ListView.__init__(
            self,
            _("Sources"),
            pdata,
            dbstate,
            uistate,
            SourceModel,
            signal_map,
            SourceBookmarks,
            nav_group,
            multiple=True,
            filter_class=SourceSidebarFilter,
        )

        self.additional_uis.append(self.additional_ui)

    def navigation_type(self):
        return "Source"

    def drag_info(self):
        return DdTargets.SOURCE_LINK

    def get_stock(self):
        return "gramps-source"

    additional_ui = [  # Defines the UI string for UIManager
        """
      <placeholder id="LocalExport">
        <item>
          <attribute name="action">win.ExportTab</attribute>
          <attribute name="label" translatable="yes">Export View...</attribute>
        </item>
      </placeholder>
""",
        """
      <section id="AddEditBook">
        <item>
          <attribute name="action">win.AddBook</attribute>
          <attribute name="label" translatable="yes">_Add Bookmark</attribute>
        </item>
        <item>
          <attribute name="action">win.EditBook</attribute>
          <attribute name="label" translatable="no">%s...</attribute>
        </item>
      </section>
"""
        % _("Organize Bookmarks"),
        """
      <placeholder id="CommonGo">
      <section>
        <item>
          <attribute name="action">win.Back</attribute>
          <attribute name="label" translatable="yes">_Back</attribute>
        </item>
        <item>
          <attribute name="action">win.Forward</attribute>
          <attribute name="label" translatable="yes">_Forward</attribute>
        </item>
      </section>
      </placeholder>
""",
        """
      <section id='CommonEdit' groups='RW'>
        <item>
          <attribute name="action">win.Add</attribute>
          <attribute name="label" translatable="yes">_Add...</attribute>
        </item>
        <item>
          <attribute name="action">win.Edit</attribute>
          <attribute name="label">%s</attribute>
        </item>
        <item>
          <attribute name="action">win.Remove</attribute>
          <attribute name="label" translatable="yes">_Delete</attribute>
        </item>
        <item>
          <attribute name="action">win.Merge</attribute>
          <attribute name="label" translatable="yes">_Merge...</attribute>
        </item>
      </section>
"""
        % _("_Edit...", "action"),  # to use sgettext()
        """
        <placeholder id='otheredit'>
        <item>
          <attribute name="action">win.FilterEdit</attribute>
          <attribute name="label" translatable="yes">"""
        """Source Filter Editor</attribute>
        </item>
        </placeholder>
""",  # Following are the Toolbar items
        """
    <placeholder id='CommonNavigation'>
    <child groups='RO'>
      <object class="GtkToolButton">
        <property name="icon-name">go-previous</property>
        <property name="action-name">win.Back</property>
        <property name="tooltip_text" translatable="yes">"""
        """Go to the previous object in the history</property>
        <property name="label" translatable="yes">_Back</property>
        <property name="use-underline">True</property>
      </object>
      <packing>
        <property name="homogeneous">False</property>
      </packing>
    </child>
    <child groups='RO'>
      <object class="GtkToolButton">
        <property name="icon-name">go-next</property>
        <property name="action-name">win.Forward</property>
        <property name="tooltip_text" translatable="yes">"""
        """Go to the next object in the history</property>
        <property name="label" translatable="yes">_Forward</property>
        <property name="use-underline">True</property>
      </object>
      <packing>
        <property name="homogeneous">False</property>
      </packing>
    </child>
    </placeholder>
""",
        """
    <placeholder id='BarCommonEdit'>
    <child groups='RW'>
      <object class="GtkToolButton">
        <property name="icon-name">list-add</property>
        <property name="action-name">win.Add</property>
        <property name="tooltip_text">%s</property>
        <property name="label" translatable="yes">_Add...</property>
        <property name="use-underline">True</property>
      </object>
      <packing>
        <property name="homogeneous">False</property>
      </packing>
    </child>
    <child groups='RW'>
      <object class="GtkToolButton">
        <property name="icon-name">gtk-edit</property>
        <property name="action-name">win.Edit</property>
        <property name="tooltip_text">%s</property>
        <property name="label" translatable="yes">Edit...</property>
        <property name="use-underline">True</property>
      </object>
      <packing>
        <property name="homogeneous">False</property>
      </packing>
    </child>
    <child groups='RW'>
      <object class="GtkToolButton">
        <property name="icon-name">list-remove</property>
        <property name="action-name">win.Remove</property>
        <property name="tooltip_text">%s</property>
        <property name="label" translatable="yes">_Delete</property>
        <property name="use-underline">True</property>
      </object>
      <packing>
        <property name="homogeneous">False</property>
      </packing>
    </child>
    <child groups='RW'>
      <object class="GtkToolButton">
        <property name="icon-name">gramps-merge</property>
        <property name="action-name">win.Merge</property>
        <property name="tooltip_text">%s</property>
        <property name="label" translatable="yes">_Merge...</property>
        <property name="use-underline">True</property>
      </object>
      <packing>
        <property name="homogeneous">False</property>
      </packing>
    </child>
    </placeholder>
"""
        % (ADD_MSG, EDIT_MSG, DEL_MSG, MERGE_MSG),
        """
    <menu id="Popup">
      <section>
        <item>
          <attribute name="action">win.Back</attribute>
          <attribute name="label" translatable="yes">_Back</attribute>
        </item>
        <item>
          <attribute name="action">win.Forward</attribute>
          <attribute name="label" translatable="yes">Forward</attribute>
        </item>
      </section>
      <section id="PopUpTree">
      </section>
      <section>
        <item>
          <attribute name="action">win.Add</attribute>
          <attribute name="label" translatable="yes">_Add...</attribute>
        </item>
        <item>
          <attribute name="action">win.Edit</attribute>
          <attribute name="label">%s</attribute>
        </item>
        <item>
          <attribute name="action">win.Remove</attribute>
          <attribute name="label" translatable="yes">_Delete</attribute>
        </item>
        <item>
          <attribute name="action">win.Merge</attribute>
          <attribute name="label" translatable="yes">_Merge...</attribute>
        </item>
      </section>
      <section>
        <placeholder id='QuickReport'>
        </placeholder>
        <placeholder id='WebConnect'>
        </placeholder>
      </section>
    </menu>
    """
        % _("_Edit...", "action"),  # to use sgettext()
    ]

    def add(self, *obj):
        EditSource(self.dbstate, self.uistate, [], Source())

    def edit(self, *obj):
        for handle in self.selected_handles():
            source = self.dbstate.db.get_source_from_handle(handle)
            try:
                EditSource(self.dbstate, self.uistate, [], source)
            except WindowActiveError:
                pass

    def merge(self, *obj):
        """
        Merge the selected sources.
        """
        mlist = self.selected_handles()

        if len(mlist) != 2:
            msg = _("Cannot merge sources.")
            msg2 = _(
                "Exactly two sources must be selected to perform a merge. "
                "A second source can be selected by holding down the "
                "control key while clicking on the desired source."
            )
            ErrorDialog(msg, msg2, parent=self.uistate.window)
        else:
            MergeSource(self.dbstate, self.uistate, [], mlist[0], mlist[1])

    def get_handle_from_gramps_id(self, gid):
        obj = self.dbstate.db.get_source_from_gramps_id(gid)
        if obj:
            return obj.get_handle()
        else:
            return None

    def tag_updated(self, handle_list):
        """
        Update tagged rows when a tag color changes.
        """
        all_links = set([])
        for tag_handle in handle_list:
            links = set(
                [
                    link[1]
                    for link in self.dbstate.db.find_backlink_handles(
                        tag_handle, include_classes="Source"
                    )
                ]
            )
            all_links = all_links.union(links)
        self.row_update(list(all_links))

    def add_tag(self, transaction, source_handle, tag_handle):
        """
        Add the given tag to the given source.
        """
        source = self.dbstate.db.get_source_from_handle(source_handle)
        source.add_tag(tag_handle)
        self.dbstate.db.commit_source(source, transaction)

    def remove_tag(self, transaction, source_handle, tag_handle):
        """
        Remove the given tag from the given source.
        """
        source = self.dbstate.db.get_source_from_handle(source_handle)
        source.remove_tag(tag_handle)
        self.dbstate.db.commit_source(source, transaction)

    def get_default_gramplets(self):
        """
        Define the default gramplets for the sidebar and bottombar.
        """
        return (
            ("Source Filter",),
            ("Source Gallery", "Source Notes", "Source Backlinks"),
        )

    def get_config_name(self):
        return __name__



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2009       Benny Malengier
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Group common stuff Gramps GUI elements must be able to do when tracking a DB:

   * connect to db signals
   * listen to db changes to update themself on relevant changes
   * determine if the GUI has become out of sync with the db
"""

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from gramps.gen.utils.callman import CallbackManager


# -------------------------------------------------------------------------
#
# DbGUIElement class
#
# -------------------------------------------------------------------------
class DbGUIElement:
    """
    Most interaction with the DB should be done via the callman attribute.
    On initialization, the method :meth:`_connect_db_signals` is called.
    Inheriting objects are advised to group the setup of the callman attribute
    here.

    .. attribute callman: a :class:`.CallbackManager` object, to be used to
                          track specific changes in the db and set up callbacks
    """

    def __init__(self, database):
        self.callman = CallbackManager(database)
        self._connect_db_signals()

    def _add_db_signal(self, name, callback):
        """
        Convenience function to add a custom db signal. The attributes are just
        passed to the callman object.
        For primary objects, use the register method of the callman attribute.

        :param name: name of the signal to connect to
        :type name: string
        :param callback: function to call when signal is emitted
        :type callback: a funtion or method with the correct signature for the
                signal
        """
        self.callman.add_db_signal(name, callback)

    def _connect_db_signals(self):
        """
        Convenience method that is called on initialization of DbGUIElement.
        Use this to group setup of the callman attribute.
        Also called in _change_db method
        """
        pass

    def _cleanup_callbacks(self):
        """
        Remove all db callbacks.
        This is done automatically on destruction of the object, but is
        normally needed earlier, calling this method does so.
        Use _change_db method if you need to remove the callbacks because the
        database has changed
        """
        database = self.callman.database
        if database.is_open():
            # a closed database has disconnected all signals
            self.callman.disconnect_all()
        # set a new callback manager
        self.callman = CallbackManager(database)

    def _change_db(self, database):
        """
        Change the database the GUI element works on to database.
        This removes all callbacks and all registered handles.

        :param database: the new database to connect to
        """
        dbold = self.callman.database
        if dbold.is_open():
            # a closed database has disconnected all signals
            self.callman.disconnect_all()
        # set a new callback manager on new database
        self.callman = CallbackManager(database)
        self._connect_db_signals()



#! /usr/bin/env python3
"""
Gramps - a GTK+/GNOME based genealogy program

Copyright (c) 2016 Gramps Development Team

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
"""

import os
import sys
import codecs
import unittest
import random

from gramps.test.test_util import Gramps
from gramps.gen.const import DATA_DIR, TEST_RANDOM
from gramps.gen.user import User
from gramps.gen.utils.id import set_det_id
from gramps.gen import const
from gramps.gen.utils.config import config

TREE_NAME = "Test_tooltest"
TEST_DIR = os.path.abspath(os.path.join(DATA_DIR, "tests"))


def call(*args):
    """Call Gramps to perform the action with out and err captured"""
    # if __debug__:
    # print ("call: %s", args)
    gramps = Gramps(user=User())
    out, err = gramps.run(*args)
    # print("out:", out, "err:", err)
    return out, err


def check_res(out, err, expected, do_out=True):
    """
    compare the output with the expected
    """
    # return all([(line in out) for line in expect])
    retval = True
    for line in expected:
        if line in (out if do_out else err):
            continue
        else:
            print("*** Expected: '%s', not found" % line)
            retval = False
    if not retval:
        if do_out:
            print(
                "*** The following was the actual stdout:\n%s"
                "*** The following was the stderr:\n%s"
                "*** End of stderr" % (out, err)
            )
        else:
            print(
                "*** The following was the actual stderr:\n%s"
                "*** The following was the stdout:\n%s"
                "*** End of stdout" % (err, out)
            )
    return retval


class ToolControl(unittest.TestCase):
    """
    These tests run some of the tools used to maintain Gramps db
    """

    def setUp(self):
        self.db_backend = config.get("database.backend")
        call("--config=database.backend:sqlite", "-y", "-q", "--remove", TREE_NAME)

    def tearDown(self):
        config.set("database.backend", self.db_backend)
        call("-y", "-q", "--remove", TREE_NAME)

    def test_tcg_and_check_and_repair(self):
        """
        Run a 'Test Case Generator' and 'Check & Repair Database' test.
        Note that the 'Test Case Generator" uses a lot of random numbers to
        generate its test cases.  This makes it less than ideal for a
        predictable unit test.  Still it contains good code for testing the
        'Check and Repair' tool.  So I force the random functions to be
        predictable by seeding it with a fixed number.  I also used the
        'Deterministic ID' function to make the usual db handle generation
        stop using random numbers and potentially reduce Gramps version to
        version issues.
        """
        # the TCG creates bad strings with illegal characters, so we need to
        # ignore them when we print the results
        try:
            sys.stderr = codecs.getwriter(sys.getdefaultencoding())(
                sys.stderr.buffer, "replace"
            )
            sys.stdout = codecs.getwriter(sys.getdefaultencoding())(
                sys.stdout.buffer, "replace"
            )
        except:
            pass
        tst_file = os.path.join(TEST_DIR, "data.gramps")
        set_det_id(True)
        TEST_RANDOM.seed(1234, version=1)
        # print(TEST_RANDOM.random())
        #         out, err = call("-s")
        #         expect = ["bsddb"]
        #         check_res(out, err, expect, do_out=True)
        out, err = call("-C", TREE_NAME, "-q", "--import", tst_file)
        expect = ["Opened successfully!", "data.gramps, format gramps.", "Cleaning up."]
        self.assertTrue(check_res(out, err, expect, do_out=False))
        self.assertEqual(out, "")
        out, err = call(
            "-O",
            TREE_NAME,
            "-y",
            "-q",
            "-a",
            "tool",
            "-p",
            "name=testcasegenerator,bugs=1,persons=1,"
            "add_linebreak=1,add_serial=1,"
            "long_names=1,lowlevel=0,person_count=50,"
            "specialchars=1",
        )
        expect = [
            "Opened successfully!",
            "Performing action: tool.",
            "Using options string: name=testcasegenerator,bugs=1",
            "Cleaning up.",
        ]
        self.assertTrue(check_res(out, err, expect, do_out=False))
        expect = ["person count 41"]
        self.assertTrue(check_res(out, err, expect, do_out=True))
        out, err = call("-O", TREE_NAME, "-y", "-a", "tool", "-p", "name=check")
        expect = [
            "1 place alternate name fixed",
            "9 media objects were referenced, but not found",
            "References to 9 missing media objects were kept",
            "3 events were referenced, but not found",
            "1 invalid birth event name was fixed",
            "1 invalid death event name was fixed",
            "2 places were referenced, but not found",
            "16 citations were referenced, but not found",
            "19 sources were referenced, but not found",
            "9 Duplicated Gramps IDs fixed",
            "9 empty objects removed:",
            "1 person objects",
            "1 family objects",
            "1 event objects",
            "1 source objects",
            "1 media objects",
            "1 place objects",
            "1 repository objects",
            "1 note objects",
        ]
        self.assertTrue(check_res(out, err, expect, do_out=True))
        expect = [
            "Opened successfully!",
            "Performing action: tool.",
            "Using options string: name=check",
            "Cleaning up.",
        ]
        self.assertTrue(check_res(out, err, expect, do_out=False))


if __name__ == "__main__":
    unittest.main()



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# test/LosHawlos_bsddbtest.py

import tempfile
from bsddb3 import db

print("Test that db.DBEnv().open() works")
with tempfile.TemporaryDirectory() as env_name:
    env = db.DBEnv()
    env.set_cachesize(0, 0x2000000)
    env.set_lk_max_locks(25000)
    env.set_lk_max_objects(25000)
    # env.set_flags(db.DB_LOG_AUTOREMOVE,1)
    """
    BSDDB change log settings using new method with renamed attributes
    """
    autoremove_flag = None
    autoremove_method = None
    for flag in ["DB_LOG_AUTO_REMOVE", "DB_LOG_AUTOREMOVE"]:
        if hasattr(db, flag):
            autoremove_flag = getattr(db, flag)
            break
    for method in ["log_set_config", "set_flags"]:
        if hasattr(env, method):
            autoremove_method = getattr(env, method)
            break
    if autoremove_method and autoremove_flag:
        autoremove_method(autoremove_flag, 1)
    else:
        print("Failed to set autoremove flag")
    env_flags = (
        db.DB_CREATE
        | db.DB_RECOVER
        | db.DB_PRIVATE
        | db.DB_INIT_MPOOL
        | db.DB_INIT_LOCK
        | db.DB_INIT_LOG
        | db.DB_INIT_TXN
        | db.DB_THREAD
    )
    try:
        env.open(env_name, env_flags)
    except db.DBRunRecoveryError as e:
        print("Exception: ")
        print(e)
        env.remove(env_name)
        env.open(env_name, env_flags)
    print("OK")



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2006  Donald N. Allingham
# Copyright (C) 2007-2008  Brian G. Matherly
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# ------------------------------------------------------------------------
#
# python modules
#
# ------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext
from gi.repository import Gtk
from gi.repository import Pango, PangoCairo

# ------------------------------------------------------------------------
#
# Gramps modules
#
# ------------------------------------------------------------------------
from gramps.gen.plug.docgen import (
    BaseDoc,
    TextDoc,
    FONT_SERIF,
    PARA_ALIGN_RIGHT,
    FONT_SANS_SERIF,
    FONT_MONOSPACE,
    PARA_ALIGN_CENTER,
    PARA_ALIGN_LEFT,
)
from ...managedwindow import ManagedWindow

RESOLUTION = PangoCairo.font_map_get_default().get_resolution()


def pixels(cm):
    return int(RESOLUTION / 2.54 * cm)


# ------------------------------------------------------------------------
#
# Constants
#
# ------------------------------------------------------------------------
LEFT, RIGHT, CENTER = "LEFT", "RIGHT", "CENTER"
_WIDTH_IN_CHARS = 72


class DisplayBuf(ManagedWindow):
    def __init__(self, title, document, track=[]):
        self.title = title
        ManagedWindow.__init__(self, document.uistate, track, document)
        dialog = Gtk.Dialog(title="", destroy_with_parent=True)
        dialog.add_button(_("_Close"), Gtk.ResponseType.CLOSE)
        self.set_window(dialog, None, title)
        self.setup_configs("interface." + title.lower().replace(" ", ""), 600, 400)
        scrolled_window = Gtk.ScrolledWindow()
        scrolled_window.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
        document.text_view = Gtk.TextView()
        document.text_view.set_buffer(document.buffer)
        self.window.connect("response", self.close)
        scrolled_window.add(document.text_view)
        self.window.vbox.pack_start(scrolled_window, True, True, 0)
        self.show()  # should use ManagedWindow version of show

    def build_menu_names(self, obj):
        return ("View", _("Quick View"))

    def get_title(self):
        return self.title


class DocumentManager:
    def __init__(self, title, document, text_view):
        self.title = title
        self.document = document
        document.text_view = text_view
        text_view.set_buffer(document.buffer)


# ------------------------------------------------------------------------
#
# TextBuf
#
# ------------------------------------------------------------------------
class TextBufDoc(BaseDoc, TextDoc):
    # --------------------------------------------------------------------
    #
    # Opens the file, resets the text buffer.
    #
    # --------------------------------------------------------------------
    def open(self, filename, container=None):
        self.has_data = True
        self.tag_table = Gtk.TextTagTable()

        sheet = self.get_style_sheet()

        for name in sheet.get_paragraph_style_names():
            tag = Gtk.TextTag(name=name)

            style = sheet.get_paragraph_style(name)
            font = style.get_font()
            if font.get_type_face() == FONT_SERIF:
                tag.set_property("family", "Serif")
            elif font.get_type_face() == FONT_SANS_SERIF:
                tag.set_property("family", "Sans")
            elif font.get_type_face() == FONT_MONOSPACE:
                tag.set_property("family", "MonoSpace")

            tag.set_property("size-points", float(font.get_size()))
            if font.get_bold():
                tag.set_property("weight", Pango.Weight.BOLD)
            if style.get_alignment() == PARA_ALIGN_RIGHT:
                tag.set_property("justification", Gtk.Justification.RIGHT)
            elif style.get_alignment() == PARA_ALIGN_LEFT:
                tag.set_property("justification", Gtk.Justification.LEFT)
            elif style.get_alignment() == PARA_ALIGN_CENTER:
                tag.set_property("justification", Gtk.Justification.CENTER)
            else:
                tag.set_property("justification", Gtk.Justification.FILL)

            if font.get_italic():
                tag.set_property("style", Pango.Style.ITALIC)

            if style.get_first_indent():
                tag.set_property("indent", pixels(style.get_first_indent()))
                # tag.set_property("tabs", [pixels(abs(style.get_first_indent()))])

            tag.set_property("left-margin", pixels(style.get_left_margin()))
            tag.set_property("right-margin", pixels(style.get_right_margin()))
            tag.set_property("pixels-above-lines", pixels(style.get_top_margin()))
            tag.set_property("pixels-below-lines", pixels(style.get_bottom_margin()))
            tag.set_property("wrap-mode", Gtk.WrapMode.WORD)

            new_tabs = style.get_tabs()

            tab_array = Pango.TabArray.new(
                initial_size=len(new_tabs) + 1, positions_in_pixels=True
            )
            index = 0
            for tab in map(pixels, new_tabs):
                tab_array.set_tab(index, Pango.TabAlign.LEFT, tab)
                index += 1
            tag.set_property("tabs", tab_array)

            self.tag_table.add(tag)
        self.buffer = Gtk.TextBuffer.new(self.tag_table)
        if container:
            return DocumentManager(_("Quick View"), self, container)
        else:
            DisplayBuf(_("Quick View"), self, track=self.track)
            return

    # --------------------------------------------------------------------
    #
    # Close the file. Call the app if required.
    #
    # --------------------------------------------------------------------
    def close(self):
        pass

    def get_usable_width(self):
        return _WIDTH_IN_CHARS

    # --------------------------------------------------------------------
    #
    # Force a section page break
    #
    # --------------------------------------------------------------------
    def end_page(self):
        return
        self.f.write("\012")

    def start_bold(self):
        pass

    def end_bold(self):
        pass

    def page_break(self):
        pass

    def start_superscript(self):
        return

    def end_superscript(self):
        return

    # --------------------------------------------------------------------
    #
    # Starts a paragraph.
    #
    # --------------------------------------------------------------------
    def start_paragraph(self, style_name, leader=None):
        self.style_name = style_name
        if leader:
            self.text = leader + "\t"
        else:
            self.text = ""

    # --------------------------------------------------------------------
    #
    # End a paragraph. First format it to the desired widths.
    # If not in table cell, write it immediately. If in the cell,
    # add it to the list for this cell after formatting.
    #
    # --------------------------------------------------------------------
    def end_paragraph(self):
        self.buffer.insert_with_tags_by_name(
            self.buffer.get_end_iter(), self.text + "\n", self.style_name
        )

    # --------------------------------------------------------------------
    #
    # Start a table. Grab the table style, and store it.
    #
    # --------------------------------------------------------------------
    def start_table(self, name, style_name):
        return
        styles = self.get_style_sheet()
        self.tbl_style = styles.get_table_style(style_name)
        self.ncols = self.tbl_style.get_columns()

    # --------------------------------------------------------------------
    #
    # End a table. Turn off the self.in_cell flag
    #
    # --------------------------------------------------------------------
    def end_table(self):
        return
        self.in_cell = 0

    # --------------------------------------------------------------------
    #
    # Start a row. Initialize lists for cell contents, number of lines,
    # and the widths. It is necessary to keep a list of cell contents
    # that is to be written after all the cells are defined.
    #
    # --------------------------------------------------------------------
    def start_row(self):
        return

        self.cellpars = [""] * self.ncols
        self.cell_lines = [0] * self.ncols
        self.cell_widths = [0] * self.ncols
        self.cellnum = -1
        self.maxlines = 0
        table_width = self.get_usable_width() * self.tbl_style.get_width() / 100.0
        for cell in range(self.ncols):
            self.cell_widths[cell] = int(
                table_width * self.tbl_style.get_column_width(cell) / 100.0
            )

    # --------------------------------------------------------------------
    #
    # End a row. Write the cell contents. Write the line of spaces
    # if the cell has fewer lines than the maximum number.
    #
    # --------------------------------------------------------------------
    def end_row(self):
        return

        self.in_cell = 0
        cell_text = [None] * self.ncols
        for cell in range(self.ncols):
            if self.cell_widths[cell]:
                blanks = " " * self.cell_widths[cell] + "\n"
                if self.cell_lines[cell] < self.maxlines:
                    self.cellpars[cell] = self.cellpars[cell] + blanks * (
                        self.maxlines - self.cell_lines[cell]
                    )
                cell_text[cell] = self.cellpars[cell].split("\n")
        for line in range(self.maxlines):
            for cell in range(self.ncols):
                if self.cell_widths[cell]:
                    self.f.write(cell_text[cell][line])
            self.f.write("\n")

    # --------------------------------------------------------------------
    #
    # Start a cell. Set the self.in_cell flag, increment the curren cell number.
    #
    # --------------------------------------------------------------------
    def start_cell(self, style_name, span=1):
        return

        self.in_cell = 1
        self.cellnum = self.cellnum + span
        span -= 1
        while span:
            self.cell_widths[self.cellnum - span] = 0
            span -= 1

    # --------------------------------------------------------------------
    #
    # End a cell. Find out the number of lines in this cell, correct
    # the maximum number of lines if necessary.
    #
    # --------------------------------------------------------------------
    def end_cell(self):
        return

        self.in_cell = 0
        self.cell_lines[self.cellnum] = self.cellpars[self.cellnum].count("\n")
        if self.cell_lines[self.cellnum] > self.maxlines:
            self.maxlines = self.cell_lines[self.cellnum]

    def add_media(self, name, align, w_cm, h_cm, alt=""):
        return

        this_text = "(photo)"
        if self.in_cell:
            self.cellpars[self.cellnum] = self.cellpars[self.cellnum] + this_text
        else:
            self.f.write(this_text)

    def write_note(self, text, format, style_name):
        return

        if format == 1:
            for line in text.split("\n"):
                self.start_paragraph(style_name)
                self.write_text(line)
                self.end_paragraph()
        elif format == 0:
            for line in text.split("\n\n"):
                self.start_paragraph(style_name)
                line = line.replace("\n", " ")
                line = " ".join(line.split())
                self.write_text(line)
                self.end_paragraph()

    # --------------------------------------------------------------------
    #
    # Writes text.
    # --------------------------------------------------------------------
    def write_text(self, text, mark=None):
        self.text = self.text + text



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2005  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
PlaceBase class for Gramps.
"""


# -------------------------------------------------------------------------
#
# PlaceBase
#
# -------------------------------------------------------------------------
class PlaceBase:
    """
    Base class for place-aware objects.
    """

    def __init__(self, source=None):
        """
        Initialize a PlaceBase.

        If the source is not None, then object is initialized from values of
        the source object.

        :param source: Object used to initialize the new object
        :type source: PlaceBase
        """
        if source:
            self.place = source.place
        else:
            self.place = ""

    def set_place_handle(self, place_handle):
        """
        Set the database handle for :class:`~.place.Place` associated with the
        object.

        :param place_handle: :class:`~.place.Place` database handle
        :type place_handle: str
        """
        self.place = place_handle

    def get_place_handle(self):
        """
        Return the database handle of the :class:`~.place.Place` associated
        with the :class:`~.event.Event`.

        :returns: :class:`~.place.Place` database handle
        :rtype: str
        """
        return self.place



# encoding:utf-8
#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2011-2016 Serge Noiraud
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# ------------------------------------------------------------------------
#
# Geography view
#
# ------------------------------------------------------------------------

# pylint: disable=bad-whitespace
# pylint: disable=bad-whitespace

MODULE_VERSION = "6.0"

from gi import Repository
from gramps.gen.const import GRAMPS_LOCALE as glocale
from gramps.gen.plug._pluginreg import register, VIEW, STABLE  # , END, START

_ = glocale.translation.gettext

# -------------------------------------------------------------------------
#
# set up logging
#
# -------------------------------------------------------------------------
import logging

_LOG = logging.getLogger("Geography")

# Attempting to import OsmGpsMap gives an error dialog if OsmGpsMap is not
# available so test first and log just a warning to the console instead.
OSMGPSMAP = False
REPOSITORY = Repository.get_default()
if REPOSITORY.enumerate_versions("OsmGpsMap"):
    try:
        # current osmgpsmap support GTK3
        import gi

        gi.require_version("OsmGpsMap", "1.0")
        from gi.repository import OsmGpsMap as osmgpsmap

        OSMGPSMAP = True
    except:
        pass

if not OSMGPSMAP:
    from gramps.gen.config import config

    if not config.get("interface.ignore-osmgpsmap"):
        from gramps.gen.constfunc import has_display

        if has_display():
            from gramps.gui.dialog import MessageHideDialog
            from gramps.gen.const import URL_WIKISTRING

            TITLE = _("OsmGpsMap module not loaded.")
            MESSAGE = _(
                "Geography functionality will not be available.\n"
                "Try to install:\n"
                " gir1.2-osmgpsmap-1.0 (debian, ubuntu, ...)\n"
                " osm-gps-map-gobject-1.0.1 for fedora, ...\n"
                " typelib-1_0-OsmGpsMap-1_0 for openSuse\n"
                " ...\n"
                "To build it for Gramps see the Wiki (<F1>)\n"
                " and search for 'build from source'"
            )
            if uistate:
                MessageHideDialog(
                    TITLE, MESSAGE, "interface.ignore-osmgpsmap", parent=uistate.window
                )
            else:
                MessageHideDialog(TITLE, MESSAGE, "interface.ignore-osmgpsmap")
else:
    # Load the view only if osmgpsmap library is present.
    register(
        VIEW,
        id="geo1",
        name=_("All known places for one Person"),
        description=_(
            "A view showing the places visited by " "one person during his life."
        ),
        version="1.0",
        gramps_target_version=MODULE_VERSION,
        status=STABLE,
        fname="geoperson.py",
        authors=["Serge Noiraud"],
        authors_email=[""],
        category=("Geography", _("Geography")),
        viewclass="GeoPerson",
        # order = START,
        stock_icon="geo-show-person",
    )

    register(
        VIEW,
        id="geo2",
        name=_("All known places for one Family"),
        description=_(
            "A view showing the places visited by " "one family during all their life."
        ),
        version="1.0",
        gramps_target_version=MODULE_VERSION,
        status=STABLE,
        fname="geofamily.py",
        authors=["Serge Noiraud"],
        authors_email=[""],
        category=("Geography", _("Geography")),
        viewclass="GeoFamily",
        # order = START,
        stock_icon="geo-show-family",
    )

    register(
        VIEW,
        id="geo3",
        name=_("Every residence or move for a person " "and any descendants"),
        description=_(
            "A view showing all the places visited by "
            "all persons during their life."
            "\nThis is for a person and any descendant."
            "\nYou can see the dates corresponding to "
            "the period."
        ),
        version="1.0",
        gramps_target_version=MODULE_VERSION,
        status=STABLE,
        fname="geomoves.py",
        authors=["Serge Noiraud"],
        authors_email=[""],
        category=("Geography", _("Geography")),
        viewclass="GeoMoves",
        # order = START,
        stock_icon="geo-show-family-down",
    )

    register(
        VIEW,
        id="geo4",
        name=_("Have these two families been able to meet?"),
        description=_(
            "A view showing the places visited by "
            "all family's members during their life: "
            "have these two people been able to meet?"
        ),
        version="1.0.1",
        gramps_target_version=MODULE_VERSION,
        status=STABLE,
        fname="geofamclose.py",
        authors=["Serge Noiraud"],
        authors_email=[""],
        category=("Geography", _("Geography")),
        viewclass="GeoFamClose",
        # order = START,
        stock_icon="geo-show-family",
    )

    register(
        VIEW,
        id="geo5",
        name=_("Have they been able to meet?"),
        description=_(
            "A view showing the places visited by "
            "two persons during their life: "
            "have these two people been able to meet?"
        ),
        version="1.0.1",
        gramps_target_version=MODULE_VERSION,
        status=STABLE,
        fname="geoclose.py",
        authors=["Serge Noiraud"],
        authors_email=[""],
        category=("Geography", _("Geography")),
        viewclass="GeoClose",
        # order = START,
        stock_icon="gramps-relation",
    )

    register(
        VIEW,
        id="geo6",
        name=_("All known Places"),
        description=_("A view showing all places of the database."),
        version="1.0",
        gramps_target_version=MODULE_VERSION,
        status=STABLE,
        fname="geoplaces.py",
        authors=["Serge Noiraud"],
        authors_email=[""],
        category=("Geography", _("Geography")),
        viewclass="GeoPlaces",
        # order = START,
        stock_icon="geo-show-place",
    )

    register(
        VIEW,
        id="geo7",
        name=_("All places related to Events"),
        description=_("A view showing all the event " "places of the database."),
        version="1.0",
        gramps_target_version=MODULE_VERSION,
        status=STABLE,
        fname="geoevents.py",
        authors=["Serge Noiraud"],
        authors_email=[""],
        category=("Geography", _("Geography")),
        viewclass="GeoEvents",
        # order = START,
        stock_icon="geo-show-event",
    )



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2010    Nick Hall
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
"""
Rule that checks for an object with a particular tag.
"""

# -------------------------------------------------------------------------
#
# Standard Python modules
#
# -------------------------------------------------------------------------
from ...const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from . import Rule


# -------------------------------------------------------------------------
#
# Typing modules
#
# -------------------------------------------------------------------------
from ...lib.primaryobj import PrimaryObject
from ...db import Database


# -------------------------------------------------------------------------
#
# HasTag
#
# -------------------------------------------------------------------------
class HasTagBase(Rule):
    """
    Rule that checks for an object with a particular tag.
    """

    labels = ["Tag:"]
    name = "Objects with the <tag>"
    description = "Matches objects with the given tag"
    category = _("General filters")

    def prepare(self, db: Database, user):
        """
        Prepare the rule. Things we want to do just once.
        """
        self.tag_handle = None
        tag = db.get_tag_from_name(self.list[0])
        if tag is not None:
            self.tag_handle = tag.handle

    def apply_to_one(self, db: Database, obj: PrimaryObject) -> bool:
        """
        Apply the rule.  Return True for a match.
        """
        if self.tag_handle is None:
            return False
        return self.tag_handle in obj.tag_list



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2002-2006  Donald N. Allingham
# Copyright (C) 2011       Tim G L Lyons
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Filter rule to match citation with a particular source.
"""
# -------------------------------------------------------------------------
#
# Standard Python modules
#
# -------------------------------------------------------------------------
from ....const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from .._hassourcebase import HasSourceBase


# -------------------------------------------------------------------------
#
# Typing modules
#
# -------------------------------------------------------------------------
from typing import Any
from ....db import Database


# -------------------------------------------------------------------------
#
# HasEvent
#
# -------------------------------------------------------------------------
class HasSource(HasSourceBase):
    """Rule that checks for an citation with a particular value"""

    labels = [_("Title:"), _("Author:"), _("Abbreviation:"), _("Publication:")]
    name = _("Sources matching parameters")
    description = _("Matches citations with a source of a particular " "value")
    category = _("Source filters")

    def apply_to_one(self, dbase: Database, citation: Any) -> bool:
        source = dbase.get_source_from_handle(citation.source_handle)
        return HasSourceBase.apply_to_one(self, dbase, source)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2006  Donald N. Allingham
# Copyright (C) 2009-2010  Nick Hall
# Copyright (C) 2009       Benny Malengier
# Copyright (C) 2010       Gary Burton
# Copyright (C) 2024       Doug Blank
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Place Model.
"""
# -------------------------------------------------------------------------
#
# python modules
#
# -------------------------------------------------------------------------
import logging

_LOG = logging.getLogger(".gui.views.treemodels.placemodel")

# -------------------------------------------------------------------------
#
# GNOME/GTK modules
#
# -------------------------------------------------------------------------
from gi.repository import Gtk

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from gramps.gen.lib import Place, PlaceType
from gramps.gen.lib.json_utils import data_to_object
from gramps.gen.datehandler import format_time
from gramps.gen.utils.place import conv_lat_lon, coord_formats
from gramps.gen.display.place import displayer as place_displayer
from gramps.gen.config import config
from .flatbasemodel import FlatBaseModel
from .treebasemodel import TreeBaseModel

# -------------------------------------------------------------------------
#
# internationalization
#
# -------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext


# -------------------------------------------------------------------------
#
# PlaceBaseModel
#
# -------------------------------------------------------------------------
class PlaceBaseModel:
    def __init__(self, db):
        self.gen_cursor = db.get_place_cursor
        self.map = db.get_raw_place_data
        self.fmap = [
            self.column_name,
            self.column_id,
            self.column_title,
            self.column_type,
            self.column_code,
            self.column_latitude,
            self.column_longitude,
            self.column_private,
            self.column_tags,
            self.column_change,
            self.column_tag_color,
            self.search_name,
        ]
        self.smap = [
            self.column_name,
            self.column_id,
            self.column_title,
            self.column_type,
            self.column_code,
            self.sort_latitude,
            self.sort_longitude,
            self.column_private,
            self.column_tags,
            self.sort_change,
            self.column_tag_color,
            self.search_name,
        ]

    def destroy(self):
        """
        Unset all elements that can prevent garbage collection
        """
        self.db = None
        self.gen_cursor = None
        self.map = None
        self.fmap = None
        self.smap = None

    def color_column(self):
        """
        Return the color column.
        """
        # This is a model.get_value() arg
        return 10

    def on_get_n_columns(self):
        return len(self.fmap) + 1

    def column_title(self, data):
        handle = data.handle
        cached, value = self.get_cached_value(handle, "PLACE")
        if not cached:
            place = data_to_object(data)
            value = place_displayer.display(self.db, place)
            self.set_cached_value(handle, "PLACE", value)
        return value

    def column_name(self, data):
        """Return the primary name"""
        return data.name.value

    def search_name(self, data):
        """The search name includes all alt names to enable finding by alt name"""
        return ",".join([data.name.value] + [name.value for name in data.alt_names])

    def column_longitude(self, data):
        if not data.long:
            return ""
        value = conv_lat_lon(
            "0",
            data.long,
            format=coord_formats[config.get("preferences.coord-format")],
        )[1]
        if not value:
            return _("Error in format")
        return ("\u202d" + value + "\u202e") if glocale.rtl_locale else value

    def column_latitude(self, data):
        if not data.lat:
            return ""
        value = conv_lat_lon(
            data.lat,
            "0",
            format=coord_formats[config.get("preferences.coord-format")],
        )[0]
        if not value:
            return _("Error in format")
        return ("\u202d" + value + "\u202e") if glocale.rtl_locale else value

    def sort_longitude(self, data):
        if not data.long:
            return ""
        value = conv_lat_lon("0", data.long, format="ISO-DMS") if data.long else ""
        if not value:
            return _("Error in format")
        return value

    def sort_latitude(self, data):
        if not data.lat:
            return ""
        value = conv_lat_lon(data.lat, "0", format="ISO-DMS") if data.lat else ""
        if not value:
            return _("Error in format")
        return value

    def column_id(self, data):
        return data.gramps_id

    def column_type(self, data):
        return PlaceType.get_str(data.place_type)

    def column_code(self, data):
        return data.code

    def column_private(self, data):
        if data.private:
            return "gramps-lock"
        else:
            # There is a problem returning None here.
            return ""

    def sort_change(self, data):
        return "%012x" % data.change

    def column_change(self, data):
        return format_time(data.change)

    def get_tag_name(self, tag_handle):
        """
        Return the tag name from the given tag handle.
        """
        cached, value = self.get_cached_value(tag_handle, "TAG_NAME")
        if not cached:
            value = self.db.get_tag_from_handle(tag_handle).get_name()
            self.set_cached_value(tag_handle, "TAG_NAME", value)
        return value

    def column_tag_color(self, data):
        """
        Return the tag color.
        """
        tag_handle = data.handle
        cached, value = self.get_cached_value(tag_handle, "TAG_COLOR")
        if not cached:
            tag_color = ""
            tag_priority = None
            for handle in data.tag_list:
                tag = self.db.get_tag_from_handle(handle)
                if tag:
                    this_priority = tag.get_priority()
                    if tag_priority is None or this_priority < tag_priority:
                        tag_color = tag.get_color()
                        tag_priority = this_priority
            value = tag_color
            self.set_cached_value(tag_handle, "TAG_COLOR", value)
        return value

    def column_tags(self, data):
        """
        Return the sorted list of tags.
        """
        tag_list = list(map(self.get_tag_name, data.tag_list))
        # TODO for Arabic, should the next line's comma be translated?
        return ", ".join(sorted(tag_list, key=glocale.sort_key))

    def clear_cache(self, handle=None):
        """
        Clear the LRU cache. Always clear lru_path, because paths may have
        changed.
        Need special version of this because Places include Placerefs, and a
        change in the place the Placeref points to can make the cache stale.
        So just clearing the effected (deleted/changed) handle is not enough.
        Note that this is only a (very) short term issue, during processing of
        signals from a merge event.  The update of the place containing a
        Placeref will occur.
        See bug 10184.

        Note that this method overrides the one from 'BaseModel', through
        FlatBaseModel and TreeBaseModel.  By locating this in PlaceBaseModel,
        the MRU rules mean this method takes precedence.
        """
        self.lru_data.clear()
        # Invalidates all paths
        self.lru_path.clear()


# -------------------------------------------------------------------------
#
# PlaceListModel
#
# -------------------------------------------------------------------------
class PlaceListModel(PlaceBaseModel, FlatBaseModel):
    """
    Flat place model.  (Original code in PlaceBaseModel).
    """

    def __init__(
        self,
        db,
        uistate,
        scol=0,
        order=Gtk.SortType.ASCENDING,
        search=None,
        skip=set(),
        sort_map=None,
    ):
        PlaceBaseModel.__init__(self, db)
        FlatBaseModel.__init__(
            self, db, uistate, scol, order, search=search, skip=skip, sort_map=sort_map
        )

    def destroy(self):
        """
        Unset all elements that can prevent garbage collection
        """
        PlaceBaseModel.destroy(self)
        FlatBaseModel.destroy(self)


# -------------------------------------------------------------------------
#
# PlaceTreeModel
#
# -------------------------------------------------------------------------
class PlaceTreeModel(PlaceBaseModel, TreeBaseModel):
    """
    Hierarchical place model.
    """

    def __init__(
        self,
        db,
        uistate,
        scol=0,
        order=Gtk.SortType.ASCENDING,
        search=None,
        skip=set(),
        sort_map=None,
    ):
        PlaceBaseModel.__init__(self, db)
        TreeBaseModel.__init__(
            self,
            db,
            uistate,
            scol=scol,
            order=order,
            search=search,
            skip=skip,
            sort_map=sort_map,
            nrgroups=3,
            group_can_have_handle=True,
        )

    def destroy(self):
        """
        Unset all elements that can prevent garbage collection
        """
        PlaceBaseModel.destroy(self)
        self.number_items = None
        TreeBaseModel.destroy(self)

    def _set_base_data(self):
        """See TreeBaseModel, for place, most have been set in init of
        PlaceBaseModel
        """
        self.number_items = self.db.get_number_of_places
        self.gen_cursor = self.db.get_place_tree_cursor

    def get_tree_levels(self):
        """
        Return the headings of the levels in the hierarchy.
        """
        return [_("Country"), _("State"), _("County"), _("Place")]

    def add_row(self, handle, data):
        """
        Add nodes to the node map for a single place.

        handle      The handle of the gramps object.
        data        The object data.
        """
        sort_key = self.sort_func(data)
        if len(data.placeref_list) > 0:
            parent = data.placeref_list[0].ref
        else:
            parent = None

        # Add the node as a root node if the parent is not in the tree.  This
        # will happen when the view is filtered.
        if not self._get_node(parent):
            parent = None

        self.add_node(parent, handle, sort_key, handle, add_parent=False)

    def column_header(self, data):
        # should not get here!
        return "????"



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2006  Donald N. Allingham
# Copyright (C) 2011       Tim G L Lyons, Nick Hall
# Copyright (C) 2024       Doug Blank
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
CitationBaseModel classes for Gramps.
"""

# -------------------------------------------------------------------------
#
# python modules
#
# -------------------------------------------------------------------------
from html import escape
import logging

log = logging.getLogger(".")
LOG = logging.getLogger(".citation")

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext
from gramps.gen.datehandler import format_time, get_date, get_date_valid
from gramps.gen.lib import Citation
from gramps.gen.lib.json_utils import data_to_object
from gramps.gen.utils.string import conf_strings
from gramps.gen.config import config

INVALID_DATE_FORMAT = config.get("preferences.invalid-date-format")


# -------------------------------------------------------------------------
#
# CitationModel
#
# -------------------------------------------------------------------------
class CitationBaseModel:
    # Fields access when 'data' is a Citation

    def citation_date(self, data):
        if data.date:
            citation = data_to_object(data)
            date_str = get_date(citation)
            if date_str != "":
                retval = escape(date_str)
            else:
                retval = ""

            if not get_date_valid(citation):
                return INVALID_DATE_FORMAT % retval
            else:
                return retval
        return ""

    def citation_sort_date(self, data):
        if data.date:
            citation = data_to_object(data)
            retval = "%09d" % citation.get_date_object().get_sort_value()
            if not get_date_valid(citation):
                return INVALID_DATE_FORMAT % retval
            else:
                return retval
        return ""

    def citation_id(self, data):
        return data.gramps_id

    def citation_page(self, data):
        return data.page

    def citation_sort_confidence(self, data):
        if data.confidence:
            return str(data.confidence)
        return ""

    def citation_confidence(self, data):
        return _(conf_strings[data.confidence])

    def citation_private(self, data):
        if data.private:
            return "gramps-lock"
        else:
            # There is a problem returning None here.
            return ""

    def citation_tags(self, data):
        """
        Return the sorted list of tags.
        """
        tag_list = list(map(self.get_tag_name, data.tag_list))
        # TODO for Arabic, should the next line's comma be translated?
        return ", ".join(sorted(tag_list, key=glocale.sort_key))

    def citation_tag_color(self, data):
        """
        Return the tag color.
        """
        tag_handle = data.handle
        cached, tag_color = self.get_cached_value(tag_handle, "TAG_COLOR")
        if not cached:
            tag_color = ""
            tag_priority = None
            for handle in data.tag_list:
                tag = self.db.get_tag_from_handle(handle)
                this_priority = tag.get_priority()
                if tag_priority is None or this_priority < tag_priority:
                    tag_color = tag.get_color()
                    tag_priority = this_priority
            self.set_cached_value(tag_handle, "TAG_COLOR", tag_color)
        return tag_color

    def citation_change(self, data):
        return format_time(data.change)

    def citation_sort_change(self, data):
        return "%012x" % data.change

    def citation_source(self, data):
        return data.source_handle

    def citation_src_title(self, data):
        source_handle = data.source_handle
        cached, value = self.get_cached_value(source_handle, "SRC_TITLE")
        if not cached:
            try:
                source = self.db.get_source_from_handle(source_handle)
                value = source.get_title()
            except:
                value = ""
            self.set_cached_value(source_handle, "SRC_TITLE", value)
        return value

    def citation_src_id(self, data):
        source_handle = data.source_handle
        cached, value = self.get_cached_value(source_handle, "SRC_ID")
        if not cached:
            try:
                source = self.db.get_source_from_handle(source_handle)
                value = source.gramps_id
            except:
                value = ""
            self.set_cached_value(source_handle, "SRC_ID", value)
        return value

    def citation_src_auth(self, data):
        source_handle = data.source_handle
        cached, value = self.get_cached_value(source_handle, "SRC_AUTH")
        if not cached:
            try:
                source = self.db.get_source_from_handle(source_handle)
                value = source.get_author()
            except:
                value = ""
            self.set_cached_value(source_handle, "SRC_AUTH", value)
        return value

    def citation_src_abbr(self, data):
        source_handle = data.source_handle
        cached, value = self.get_cached_value(source_handle, "SRC_ABBR")
        if not cached:
            try:
                source = self.db.get_source_from_handle(source_handle)
                value = source.get_abbreviation()
            except:
                value = ""
            self.set_cached_value(source_handle, "SRC_ABBR", value)
        return value

    def citation_src_pinfo(self, data):
        source_handle = data.source_handle
        cached, value = self.get_cached_value(source_handle, "SRC_PINFO")
        if not cached:
            try:
                source = self.db.get_source_from_handle(source_handle)
                value = source.get_publication_info()
            except:
                value = ""
            self.set_cached_value(source_handle, "SRC_PINFO", value)
        return value

    def citation_src_private(self, data):
        source_handle = data.source_handle
        cached, value = self.get_cached_value(source_handle, "SRC_PRIVATE")
        if not cached:
            try:
                source = self.db.get_source_from_handle(source_handle)
                if source.get_privacy():
                    value = "gramps-lock"
                else:
                    # There is a problem returning None here.
                    value = ""
            except:
                value = ""
            self.set_cached_value(source_handle, "SRC_PRIVATE", value)
        return value

    def citation_src_tags(self, data):
        source_handle = data.source_handle
        cached, value = self.get_cached_value(source_handle, "SRC_TAGS")
        if not cached:
            try:
                source = self.db.get_source_from_handle(source_handle)
                tag_list = list(map(self.get_tag_name, source.get_tag_list()))
                # TODO for Arabic, should the next line's comma be translated?
                value = ", ".join(sorted(tag_list, key=glocale.sort_key))
            except:
                value = ""
            self.set_cached_value(source_handle, "SRC_TAGS", value)
        return value

    def citation_src_chan(self, data):
        source_handle = data.source_handle
        cached, value = self.get_cached_value(source_handle, "SRC_CHAN")
        if not cached:
            try:
                source = self.db.get_source_from_handle(source_handle)
                value = format_time(source.change)
            except:
                value = ""
            self.set_cached_value(source_handle, "SRC_CHAN", value)
        return value

    def citation_src_sort_change(self, data):
        source_handle = data.source_handle
        cached, value = self.get_cached_value(source_handle, "SRC_CHAN")
        if not cached:
            try:
                source = self.db.get_source_from_handle(source_handle)
                value = "%012x" % source.change
            except:
                value = ""
            self.set_cached_value(source_handle, "SRC_CHAN", value)
        return value

    # Fields access when 'data' is a Source

    def source_src_title(self, data):
        return data.title

    def source_src_id(self, data):
        return data.gramps_id

    def source_src_auth(self, data):
        return data.author

    def source_src_abbr(self, data):
        return data.abbrev

    def source_src_pinfo(self, data):
        return data.pubinfo

    def source_src_private(self, data):
        if data.private:
            return "gramps-lock"
        else:
            # There is a problem returning None here.
            return ""

    def source_src_tags(self, data):
        """
        Return the sorted list of tags.
        """
        tag_list = list(map(self.get_tag_name, data.tag_list))
        # TODO for Arabic, should the next line's comma be translated?
        return ", ".join(sorted(tag_list, key=glocale.sort_key))

    def source_src_tag_color(self, data):
        """
        Return the tag color.
        """
        tag_handle = data.handle
        cached, tag_color = self.get_cached_value(tag_handle, "TAG_COLOR")
        if not cached:
            tag_color = ""
            tag_priority = None
            for handle in data.tag_list:
                tag = self.db.get_tag_from_handle(handle)
                this_priority = tag.get_priority()
                if tag_priority is None or this_priority < tag_priority:
                    tag_color = tag.get_color()
                    tag_priority = this_priority
            self.set_cached_value(tag_handle, "TAG_COLOR", tag_color)
        return tag_color

    def source_src_chan(self, data):
        return format_time(data.change)

    def source_sort2_change(self, data):
        return "%012x" % data.change

    def dummy_sort_key(self, data):
        # dummy sort key for columns that don't have data
        return None

    def get_tag_name(self, tag_handle):
        """
        Return the tag name from the given tag handle.
        """
        cached, value = self.get_cached_value(tag_handle, "TAG_NAME")
        if not cached:
            value = self.db.get_tag_from_handle(tag_handle).get_name()
            self.set_cached_value(tag_handle, "TAG_NAME", value)
        return value



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2015 Douglas Blank <doug.blank@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
from gramps.gen.plug._pluginreg import newplugin, STABLE, DATABASE
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext

plg = newplugin()
plg.id = "bsddb"
plg.name = _("BSDDB")
plg.name_accell = _("_BSDDB Database")
plg.description = _("Berkeley Software Distribution Database Backend")
plg.version = "2.0"
plg.gramps_target_version = "6.0"
plg.status = STABLE
plg.fname = "bsddb.py"
plg.ptype = DATABASE
plg.databaseclass = "DbBsddb"



#! /usr/bin/env python
#
# update_po - a gramps tool to update translations
#
# Copyright (C) 2006-2006  Kees Bakker
# Copyright (C) 2006       Brian Matherly
# Copyright (C) 2008       Stephen George
# Copyright (C) 2012
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

"""
update_man.py for command line documentation.

Examples:
   python update_man.py -t

      Tests if 'sphinx' and 'python' are well configured.
"""

from __future__ import print_function

import os
import sys
from argparse import ArgumentParser

DOCUTILS = True
try:
    import docutils.core, docutils.writers
except:
    DOCUTILS = False

LANGUAGES = ["sv", "nl", "pl", "cs", "pt_BR", "fr"]
VERSION = "5.0.0"
DATE = ""

# You can set these variables from the command line.
SPHINXBUILD = "sphinx-build"

if sys.platform == "win32":
    pythonCmd = os.path.join(sys.prefix, "bin", "python.exe")
    sphinxCmd = os.path.join(sys.prefix, "bin", "sphinx-build.exe")
elif sys.platform in ["linux2", "darwin", "cygwin"]:
    pythonCmd = os.path.join(sys.prefix, "bin", "python")
    sphinxCmd = SPHINXBUILD
else:
    print("Update Man ERROR: unknown system, don't know sphinx, ... commands")
    sys.exit(0)


def tests():
    """
    Testing installed programs.
    We made tests (-t flag) by displaying versions of tools if properly
    installed. Cannot run all commands without 'sphinx' and 'python'.
    """
    try:
        print("\n=================='python'=============================\n")
        os.system("""%(program)s -V""" % {"program": pythonCmd})
    except:
        print("Please, install python")

    try:
        print("\n=================='Sphinx-build'=============================\n")
        os.system("""%(program)s""" % {"program": sphinxCmd})
    except:
        print("Please, install sphinx")

    if not DOCUTILS:
        print("\nNo docutils support, cannot use -m/--man and -o/--odt arguments.")


def main():
    """
    The utility for handling documentation stuff.
    What is need by Gramps, nothing more.
    """

    parser = ArgumentParser(
        description="This program aims to handle documentation"
        " and related translated versions.",
    )

    parser.add_argument(
        "-t",
        "--test",
        action="store_true",
        dest="test",
        default=True,
        help="test if 'python' and 'sphinx' are properly installed",
    )

    parser.add_argument(
        "-b",
        "--build",
        action="store_true",
        dest="build",
        default=False,
        help="build man documentation (via sphinx-build)",
    )

    parser.add_argument(
        "-m",
        "--man",
        action="store_true",
        dest="man",
        default=False,
        help="build man documentation (via docutils)",
    )

    parser.add_argument(
        "-o",
        "--odt",
        action="store_true",
        dest="odt",
        default=False,
        help="build odt documentation (via docutils)",
    )

    args = parser.parse_args()

    if args.test:
        tests()

    if args.build:
        build()

    if args.man and DOCUTILS:
        man()

    if args.odt and DOCUTILS:
        odt()


def build():
    """
    Build documentation.
    """

    # testing stage

    os.system("""%(program)s -b html . _build/html""" % {"program": sphinxCmd})
    os.system("""%(program)s -b htmlhelp . _build/htmlhelp""" % {"program": sphinxCmd})
    if DOCUTILS:
        os.system("""%(program)s -b man . .""" % {"program": sphinxCmd})
    os.system("""%(program)s -b text . _build/text""" % {"program": sphinxCmd})
    os.system("""%(program)s -b changes . _build/changes""" % {"program": sphinxCmd})
    # os.system('''%(program)s -b linkcheck . _build/linkcheck''' % {'program': sphinxCmd})
    os.system("""%(program)s -b gettext . _build/gettext""" % {"program": sphinxCmd})

    for lang in LANGUAGES:
        os.system(
            """%(program)s -b html -D language="%(lang)s" master_doc="%(lang)s" %(lang)s %(lang)s"""
            % {"lang": lang, "program": sphinxCmd}
        )
        os.system(
            """%(program)s -b htmlhelp -D language="%(lang)s" master_doc="%(lang)s" %(lang)s %(lang)s"""
            % {"lang": lang, "program": sphinxCmd}
        )
        if DOCUTILS:
            os.system(
                """%(program)s -b man %(lang)s %(lang)s"""
                % {"lang": lang, "program": sphinxCmd}
            )
        os.system(
            """%(program)s -b text -D language="%(lang)s" master_doc="%(lang)s" %(lang)s %(lang)s"""
            % {"lang": lang, "program": sphinxCmd}
        )
        # for update/migration
        os.system(
            """%(program)s -b gettext -D language="%(lang)s" master_doc="%(lang)s" . _build/gettext/%(lang)s"""
            % {"lang": lang, "program": sphinxCmd}
        )


def man():
    """
    man file generation via docutils (python)

    from docutils.core import publish_cmdline, default_description
    from docutils.writers import manpage
    """

    os.system("""rst2man en.rst gramps.1""")

    for lang in LANGUAGES:
        os.system(
            """rst2man %(lang)s/%(lang)s.rst -l %(lang)s %(lang)s/gramps.1"""
            % {"lang": lang}
        )


def odt():
    """
    odt file generation via docutils (python)

    from docutils.core import publish_cmdline_to_binary, default_description
    from docutils.writers.odf_odt import Writer, Reader
    """

    os.system("""rst2odt en.rst gramps.odt""")

    for lang in LANGUAGES:
        os.system(
            """rst2odt %(lang)s/%(lang)s.rst -l %(lang)s %(lang)s/gramps.odt"""
            % {"lang": lang}
        )


if __name__ == "__main__":
    main()



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""CLI tests for gramps"""

import sys
import os
import unittest
import re
import subprocess

from gramps.gen.dbstate import DbState
from gramps.test.test_util import Gramps

test_ged = """0 HEAD
1 SOUR min1r.ged min 1-rec
1 SUBM @SUBM1@
1 GEDC
2 VERS 5.5
2 FORM LINEAGE-LINKED
1 CHAR ASCII
0 @I1@ INDI
0 TRLR
"""

ddir = os.path.dirname(__file__)
min1r = os.path.join(ddir, "min1r.ged")
out_ged = os.path.join(ddir, "test_out.ged")
example_copy = os.path.join(ddir, "copy.gramps")
example = os.path.join(ddir, "..", "..", "..", "example", "gramps", "data.gramps")


class Test(unittest.TestCase):
    def setUp(self):
        self.tearDown()

    def tearDown(self):
        if os.path.exists(out_ged):
            os.remove(out_ged)

    @classmethod
    def setUpClass(cls):
        if not os.path.exists(min1r):
            with open(min1r, "w") as f:
                f.write(test_ged)

    @classmethod
    def tearDownClass(cls):
        if os.path.exists(min1r):
            os.remove(min1r)

    # silly test just to illustrate unittest setUp behavior
    def test1_setup_works(self):
        self.assertTrue(os.path.exists(ddir), "data dir %r exists" % ddir)
        self.assertTrue(os.path.exists(min1r), "data file %r exists" % min1r)
        self.assertFalse(os.path.exists(out_ged), "NO out file %r yet" % out_ged)

    # This tests the fix for bug #1331-1334
    # read trivial gedcom input, write gedcom output
    def test2_exec_CLI(self):
        ifile = min1r
        ofile = out_ged
        gcmd = [sys.executable, "Gramps.py", "-i", ifile, "-e", ofile]
        process = subprocess.Popen(
            gcmd, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE
        )
        result_str, err_str = process.communicate()
        self.assertEqual(process.returncode, 0, "executed CLI command %r" % gcmd)
        # simple validation o output
        self.assertTrue(os.path.isfile(ofile), "output file created")
        with open(ofile) as f:
            content = f.read()
        g = re.search("INDI", content)
        self.assertTrue(g, "found 'INDI' in output file")

    def test2_exec_cli_m(self):
        """Test the `python -m gramps` way to run the CLI."""
        ifile = min1r
        ofile = out_ged
        gcmd = [sys.executable, "-m", "gramps", "-i", ifile, "-e", ofile]
        process = subprocess.Popen(
            gcmd, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE
        )
        result_str, err_str = process.communicate()
        self.assertEqual(process.returncode, 0, "executed CLI command %r" % gcmd)
        # simple validation o output
        self.assertTrue(os.path.isfile(ofile), "output file created")
        with open(ofile) as f:
            content = f.read()
        g = re.search("INDI", content)
        self.assertTrue(g, "found 'INDI' in output file")


class UnicodeTest(unittest.TestCase):
    def setUp(self):
        from gramps.cli.clidbman import CLIDbManager
        from gramps.gen.config import set as setconfig, get as getconfig
        from gramps.gen.dbstate import DbState

        self.newpath = os.path.join(
            os.path.dirname(__file__), "\u0393\u03c1\u03b1\u03bc\u03c3\u03c0"
        )
        self.newtitle = "Gr\u00e4mps T\u00e9st"
        os.makedirs(self.newpath)
        self.old_path = getconfig("database.path")
        setconfig("database.path", self.newpath)
        self.cli = CLIDbManager(DbState())

    def tearDown(self):
        from gramps.gen.config import set as setconfig

        for dirpath, dirnames, filenames in os.walk(self.newpath, False):
            for afile in filenames:
                os.remove(os.path.join(dirpath, afile))
            for adir in dirnames:
                os.rmdir(os.path.join(dirpath, adir))
        os.rmdir(self.newpath)
        setconfig("database.path", self.old_path)

    # Test that clidbman will open files in a path containing
    # arbitrary Unicode characters.
    def test4_arbitrary_uncode_path(self):
        (dbpath, title) = self.cli.create_new_db_cli(self.newtitle)

        self.assertEqual(
            self.newpath,
            os.path.dirname(dbpath),
            "Compare paths %s and %s" % (repr(self.newpath), repr(dbpath)),
        )
        self.assertEqual(
            self.newtitle,
            title,
            "Compare titles %s and %s" % (repr(self.newtitle), repr(title)),
        )


class CLITest(unittest.TestCase):
    def tearDown(self):
        if os.path.exists(example_copy):
            os.remove(example_copy)
        self.call("--remove", "Test: test1_cli")

    def setUp(self):
        self.gramps = Gramps()
        self.tearDown()

    def call(self, *args, stdin=None):
        self.gramps.run(*args, stdin=stdin)

    def test1a_cli(self):
        self.call("-C", "Test: test1_cli", "--import", example)

    def test1b_cli(self):
        self.call("-O", "Test: test1_cli", "--export", example_copy)


if __name__ == "__main__":
    unittest.main()

# ===eof===
